import threading

from visidata import vd, UNLOADED, namedlist, vlen, asyncthread, globalCommand, date
from visidata import VisiData, BaseSheet, Sheet, ColumnAttr, VisiDataMetaSheet, JsonLinesSheet, TypedWrapper, AttrDict, Progress, ErrorSheet, CompleteKey, Path, ColumnsSheet
import visidata

vd.option('replay_wait', 0.0, 'time to wait between replayed commands, in seconds', sheettype=None)
vd.option('replay_ignore_errors', False, 'continue replay on error instead of aborting', sheettype=None)
vd.theme_option('disp_replay_play', '▶', 'status indicator for active replay')
vd.theme_option('disp_replay_record', '⏺', 'status indicator for macro record')
vd.theme_option('color_status_replay', 'green', 'color of replay status indicator')

# prefixes which should not be logged
nonLogged = '''forget exec-longname undo redo quit
show error errors statuses options threads jump
replay cancel save-cmdlog macro cmdlog-sheet menu repeat reload-every
search scroll prev next page start end zoom visibility sidebar
mouse suspend redraw no-op help syscopy sysopen profile toggle'''.split()

vd.option('rowkey_prefix', 'キ', 'string prefix for rowkey in the cmdlog', sheettype=None)

vd._nextCommands = []  # list[str|CommandLogRow] for vd.queueCommand

CommandLogRow = namedlist('CommandLogRow', 'sheet col row longname input keystrokes comment undofuncs'.split())

@VisiData.api
def queueCommand(vd, longname, input=None, sheet=None, col=None, row=None):
    'Add command to queue of next commands to execute.'
    vd._nextCommands.append(CommandLogRow(longname=longname, input=input, sheet=sheet, col=col, row=row))


@VisiData.api
def open_vd(vd, p):
    return CommandLog(p.base_stem, source=p, precious=True)

@VisiData.api
def open_vdj(vd, p):
    return CommandLogJsonl(p.base_stem, source=p, precious=True)

VisiData.save_vd = VisiData.save_tsv


@VisiData.api
def save_vdj(vd, p, *vsheets):
    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        fp.write("#!/usr/bin/env -S vd -p\n")
        fp.write(f"# {visidata.__version_info__}\n")
        for vs in vsheets:
            vs.write_jsonl(fp)


@VisiData.api
def checkVersion(vd, desired_version):
    if desired_version != visidata.__version_info__:
        vd.fail("version %s required" % desired_version)

@VisiData.api
def fnSuffix(vd, prefix:str):
    i = 0
    fn = prefix + '.vdj'
    while Path(fn).exists():
        i += 1
        fn = f'{prefix}-{i}.vdj'

    return fn

def indexMatch(L, func):
    'returns the smallest i for which func(L[i]) is true'
    for i, x in enumerate(L):
        if func(x):
            return i

@VisiData.api
def isLoggableCommand(vd, cmd):
    'Return whether command should be logged to the cmdlog, depending if it has a prefix in nonLogged, or was defined with replay=False.'
    if not cmd.replayable:
        return False

    for n in nonLogged:
        if cmd.longname.startswith(n):
            return False
    return True

def isLoggableSheet(sheet):
    return sheet is not vd.cmdlog and not isinstance(sheet, (vd.OptionsSheet, ErrorSheet))


@Sheet.api
def moveToRow(vs, rowstr):
    'Move cursor to row given by *rowstr*, which can be either the row number or keystr.'
    rowidx = vs.getRowIndexFromStr(rowstr)
    if rowidx is None or rowidx >= vs.nRows:
        return False

    vs.cursorRowIndex = rowidx

    return True

@Sheet.api
def getRowIndexFromStr(vs, row):
    prefix = vd.options.rowkey_prefix
    index = None
    if isinstance(row, int):
        index = row
    elif isinstance(row, str) and row.startswith(prefix):
        rowk = row[len(prefix):]
        index = indexMatch(vs.rows, lambda r,vs=vs,rowk=rowk: rowk == ','.join(map(str, vs.rowkey(r))))
    else:
        try:
            index = int(row)
        except ValueError:
            vd.warning('invalid type for row index')

    return index


@Sheet.api
def moveToCol(vs, col):
    'Move cursor to column given by *col*, which can be either the column number or column name.'
    vcolidx = None
    if isinstance(col, str):
        vcolidx = indexMatch(vs.availCols, lambda c,name=col: name == c.name)
    elif isinstance(col, int):
        vcolidx = col

    if vcolidx is None or vcolidx >= len(vs.availCols):
        return False

    vs.cursorVisibleColIndex = vcolidx

    return True


@BaseSheet.api
def commandCursor(sheet, execstr):
    'Return (col, row) of cursor suitable for cmdlog replay of execstr.'
    colname, rowname = '', ''
    contains = lambda s, *substrs: any((a in s) for a in substrs)
    if contains(execstr, 'cursorTypedValue', 'cursorDisplay', 'cursorValue', 'cursorCell', 'cursorRow') and sheet.nRows > 0:
        rowname = sheet.cursorRowIndex

    if contains(execstr, 'cursorTypedValue', 'cursorDisplay', 'cursorValue', 'cursorCell', 'cursorCol', 'cursorVisibleCol', 'ColumnAtCursor'):
        if sheet.cursorCol:
            colname = sheet.cursorCol.name or sheet.visibleCols.index(sheet.cursorCol)
        else:
            colname = None
    return colname, rowname


# rowdef: namedlist (like TsvSheet)
class CommandLogBase:
    'Log of commands for current session.'
    rowtype = 'logged commands'
    precious = False
    _rowtype = CommandLogRow
    columns = [
        ColumnAttr('sheet'),
        ColumnAttr('col'),
        ColumnAttr('row'),
        ColumnAttr('longname'),
        ColumnAttr('input'),
        ColumnAttr('keystrokes'),
        ColumnAttr('comment'),
        ColumnAttr('undo', 'undofuncs', type=vlen, width=0)
    ]

    filetype = 'vd'

    def newRow(self, **fields):
        return self._rowtype(**fields)

    def beforeExecHook(self, sheet, cmd, args, keystrokes):
        if vd.activeCommand:
            self.afterExecSheet(sheet, False, '')

        colname, rowname, sheetname = '', '', None
        if sheet and not (cmd.longname.startswith('open-') and cmd.longname not in ('open-row', 'open-cell')):
            sheetname = sheet.name

            colname, rowname = sheet.commandCursor(cmd.execstr)

            contains = lambda s, *substrs: any((a in s) for a in substrs)
            if contains(cmd.execstr, 'pasteFromClipboard'):
                args = vd.sysclipValue().strip()


        comment = vd.currentReplayRow.comment if vd.currentReplayRow else cmd.helpstr
        vd.activeCommand = self.newRow(sheet=sheetname,
                                            col=colname,
                                            row=rowname,
                                            keystrokes=keystrokes,
                                            input=args,
                                            longname=cmd.longname,
                                            comment=comment,
                                            replayable=cmd.replayable,
                                            undofuncs=[])

    def afterExecSheet(self, sheet, escaped, err):
        'Records vd.activeCommand'
        if not vd.activeCommand:  # nothing to record
            return

        if escaped:
            vd.activeCommand = None
            return

        # remove user-aborted commands and simple movements (unless first command on the sheet, which created the sheet)
        if not sheet.cmdlog_sheet.rows or vd.isLoggableCommand(vd.activeCommand):
            if isLoggableSheet(sheet):      # don't record actions from cmdlog or other internal sheets on global cmdlog
                self.addRow(vd.activeCommand)  # add to global cmdlog
            sheet.cmdlog_sheet.addRow(vd.activeCommand)  # add to sheet-specific cmdlog

        vd.activeCommand = None

    def openHook(self, vs, src):
        while isinstance(src, BaseSheet):
            src = src.source
        r = self.newRow(keystrokes='o', input=str(src), longname='open-file', replayable=True)
        vs.cmdlog_sheet.addRow(r)
        self.addRow(r)

class CommandLog(CommandLogBase, VisiDataMetaSheet):
    pass

class CommandLogJsonl(CommandLogBase, JsonLinesSheet):

    filetype = 'vdj'

    def newRow(self, **fields):
        return AttrDict(JsonLinesSheet.newRow(self, **fields))

    def iterload(self):
        for r in JsonLinesSheet.iterload(self):
            if isinstance(r, TypedWrapper):
                yield r
            else:
                yield AttrDict(r)


### replay

vd.paused = False
vd.currentReplay = None     # CommandLog replaying currently
vd.currentReplayRow = None  # must be global, to allow replay


@VisiData.api
def replay_cancel(vd):
    vd.currentReplayRow = None
    vd.currentReplay = None
    vd._nextCommands.clear()


@VisiData.api
def moveToReplayContext(vd, r, vs):
        'set the sheet/row/col to the values in the replay row'
        vs.ensureLoaded()
        vd.sync()
        vd.clearCaches()

        if r.row not in [None, '']:
            vs.moveToRow(r.row) or vd.error(f'no `{r.row}` row on {vs}')

        if r.col not in [None, '']:
            vs.moveToCol(r.col) or vd.error(f'no `{r.col}` column on {vs}')


@VisiData.api
def replayOne(vd, r):
        'Replay the command in one given row.'
        vd.currentReplayRow = r
        longname = getattr(r, 'longname', None)
        if longname is None and getattr(r, 'keystrokes', None) is None:
            vd.fail('failed to find command to replay')

        if r.sheet and longname not in ['set-option', 'unset-option']:
            vs = vd.getSheet(r.sheet) or vd.error(f'no sheet named `{r.sheet}`')
        else:
            vs = None

        if longname in ['set-option', 'unset-option']:
            try:
                context = vs if r.sheet and vs else vd
                option_scope = r.sheet or r.col or 'global'
                if option_scope == 'override': option_scope = 'global' # override is deprecated, is now global
                if longname == 'set-option':
                    context.options.set(r.row, r.input, option_scope)
                else:
                    context.options.unset(r.row, option_scope)

                escaped = False
            except Exception as e:
                vd.exceptionCaught(e)
                escaped = True
        else:
            vs = vs or vd.activeSheet
            if vs:
                if vs in vd.sheets:  # if already on sheet stack, push to top
                    vd.push(vs)
            else:
                vs = vd.cmdlog

            try:
                vd.moveToReplayContext(r, vs)
                if r.comment:
                    vd.status(r.comment)

                # <=v1.2 used keystrokes in longname column; getCommand fetches both
                escaped = vs.execCommand(longname if longname else r.keystrokes, keystrokes=r.keystrokes)
            except Exception as e:
                vd.exceptionCaught(e)
                escaped = True

        vd.currentReplayRow = None

        if escaped:  # escape during replay aborts replay
            vd.warning('replay aborted during %s' % (longname or r.keystrokes))
        return escaped


@VisiData.api
class DisableAsync:
    def __enter__(self):
        vd.execAsync = vd.execSync

    def __exit__(self, exc_type, exc_val, tb):
        vd.execAsync = lambda *args, vd=vd, **kwargs: visidata.VisiData.execAsync(vd, *args, **kwargs)


@VisiData.api
def replay_sync(vd, cmdlog):
    'Replay all commands in *cmdlog*.'
    with vd.DisableAsync():
        vd.sync()  #2352 let cmdlog finish loading
        cmdlog.cursorRowIndex = 0
        # save current replay, for cmdlogs that replay other cmdlogs, such as a macro executing another macro
        prev_replay = vd.currentReplay
        prev_replay_row = vd.currentReplayRow
        vd.currentReplay = cmdlog

        with Progress(total=len(cmdlog.rows)) as prog:
            while cmdlog.cursorRowIndex < len(cmdlog.rows):
                if vd.currentReplay is None:
                    vd.status('replay canceled')
                    return

                vd.statuses.clear()
                try:
                    if vd.replayOne(cmdlog.cursorRow):
                        if not vd.options.replay_ignore_errors:
                            vd.replay_cancel()
                            return True
                except Exception as e:
                    vd.exceptionCaught(e)
                    if not vd.options.replay_ignore_errors:
                        vd.replay_cancel()
                        vd.status('replay canceled')
                        return True

                cmdlog.cursorRowIndex += 1
                prog.addProgress(1)

                if vd.activeSheet:
                    vd.activeSheet.ensureLoaded()

        vd.status('replay complete')
        vd.currentReplay = prev_replay
        vd.currentReplayRow = prev_replay_row


@VisiData.api
def replay(vd, cmdlog):
    'Inject commands into live execution with interface.'
    vd._nextCommands.extend(cmdlog.rows)
    vd.currentReplay = cmdlog


@VisiData.api
def getLastArgs(vd):
        'Get user input for the currently playing command.'
        if vd.currentReplayRow:
            return vd.currentReplayRow.input
        return None


@VisiData.api
def setLastArgs(vd, args):
        'Set user input on last command, if not already set.'
        # only set if not already set (second input usually confirmation)
        if vd.activeCommand:
            if not vd.activeCommand.input:
                vd.activeCommand.input = args


@VisiData.property
def replayStatus(vd):
    if vd.macroMode:
        return f'|[:error] {len(vd.macroMode)} {vd.options.disp_replay_record} [:]'

    if vd._nextCommands:
        return f'|[:status_replay] {len(vd._nextCommands)} {vd.options.disp_replay_play} [:]'

    return ''


@BaseSheet.property
def cmdlog(sheet):
    rows = sheet.cmdlog_sheet.rows
    if isinstance(sheet.source, BaseSheet):
        rows = sheet.source.cmdlog.rows + rows
    return CommandLogJsonl(sheet.name+'_cmdlog', source=sheet, rows=rows)


@BaseSheet.lazy_property
def cmdlog_sheet(sheet):
    c = CommandLogJsonl(sheet.name+'_cmdlog', source=sheet, rows=[])
    # copy over all existing globally set options
    # you only need to do this for the first BaseSheet in a tree
    if not isinstance(sheet.source, BaseSheet):
        for r in vd.cmdlog.rows:
            if r.sheet == 'global' and (r.longname == 'set-option') or (r.longname == 'unset-option'):
                c.addRow(r)
    return c


@ColumnsSheet.property
def cmdlog_sheet(sheet):
    'Edits to a single source sheet log (and undo) against that source. #3133'
    if len(sheet.source) == 1 and isinstance(sheet.source[0], BaseSheet):
        return sheet.source[0].cmdlog_sheet
    return super(ColumnsSheet, sheet).cmdlog_sheet


@BaseSheet.property
def shortcut(self):
    if self._shortcut:
        return self._shortcut
    try:
        return str(vd.allSheets.index(self)+1)
    except ValueError:
        pass

    try:
        return self.cmdlog_sheet.rows[0].keystrokes or ''  #2293
    except Exception:
        pass

    return ''


@VisiData.property
def cmdlog(vd):
    if not vd._cmdlog:
        vd._cmdlog = CommandLogJsonl('cmdlog', rows=[])  # no reload
        with vd.suppressUndo():  # building a sheet's layout is not an undoable user action
            vd._cmdlog.resetCols()
        vd.beforeExecHooks.append(vd._cmdlog.beforeExecHook)
    return vd._cmdlog

@VisiData.property
def modifyCommand(vd):
    if vd.activeCommand and vd.isLoggableCommand(vd.activeCommand):
        return vd.activeCommand
    if not vd.cmdlog.rows:
        return None
    return vd.cmdlog.rows[-1]


BaseSheet.init('_shortcut')


globalCommand('gD', 'cmdlog-all', 'vd.push(vd.cmdlog)', 'open global CommandLog for all commands executed in current session')
globalCommand('D', 'cmdlog-sheet', 'vd.push(sheet.cmdlog)', "open current sheet's CommandLog with all other loose ends removed; includes commands from parent sheets")
globalCommand('zD', 'cmdlog-sheet-only', 'vd.push(sheet.cmdlog_sheet)', 'open CommandLog for current sheet with commands from parent sheets removed')
BaseSheet.addCommand('Ctrl+D', 'save-cmdlog', 'saveSheets(inputPath("save cmdlog to: ", value=fnSuffix(name)), vd.cmdlog)', 'save CommandLog to filename.vdj file')
BaseSheet.bindkey('Ctrl+N', 'no-op')
BaseSheet.addCommand('Ctrl+K', 'replay-stop', 'vd.replay_cancel(); vd.warning("replay canceled")', 'cancel current replay')

globalCommand(None, 'show-status', 'status(input("status: "))', 'show given message on status line')
globalCommand('Ctrl+V', 'show-version', 'status(__version_info__);', 'Show version and copyright information on status line')
globalCommand('zCtrl+V', 'check-version', 'checkVersion(input("require version: ", value=__version_info__))', 'check VisiData version against given version')

CommandLog.addCommand('x', 'replay-row', 'vd.replayOne(cursorRow); status("replayed one row")', 'replay command in current row')
CommandLog.addCommand('gx', 'replay-all', 'vd.replay(sheet)', 'replay contents of entire CommandLog')

CommandLogJsonl.addCommand('x', 'replay-row', 'vd.replayOne(cursorRow); status("replayed one row")', 'replay command in current row')
CommandLogJsonl.addCommand('gx', 'replay-all', 'vd.replay(sheet)', 'replay contents of entire CommandLog')

CommandLog.options.json_sort_keys = False
CommandLog.options.encoding = 'utf-8'
CommandLogJsonl.options.json_sort_keys = False
CommandLogJsonl.options.regex_skip = r'^(//|#).*'

vd.addGlobals(CommandLogBase=CommandLogBase,
              CommandLogJsonl=CommandLogJsonl,
              CommandLogRow=CommandLogRow,
              )

vd.addMenuItems('''
    View > Command log > this sheet > cmdlog-sheet
    View > Command log > this sheet only > cmdlog-sheet-only
    View > Command log > all commands > cmdlog-all
    System > Execute longname > exec-longname
    Help > Version > show-version
''')



import itertools
import collections
import math
import time

from visidata import vd, asyncthread, options, Progress, ColumnItem, SequenceSheet, Sheet, VisiData
from visidata import namedlist, filesize

vd.option('delimiter', '\t', 'field delimiter to use for tsv/csv filetype', replay=True)
vd.option('row_delimiter', '\n', 'row delimiter to use for tsv/csv filetype', replay=True)
vd.option('tsv_safe_newline', '\u001e', 'replacement for newline character when saving to tsv', replay=True)
vd.option('tsv_safe_tab', '\u001f', 'replacement for tab character when saving to tsv', replay=True)


@VisiData.api
def open_tsv(vd, p):
    return TsvSheet(p.base_stem, source=p)


def adaptive_bufferer(fp, max_buffer_size=65536):
    """Loading e.g. tsv files goes faster with a large buffer. But when the input stream
    is slow (e.g. 1 byte/second) and the buffer size is large, it can take a long time until
    the buffer is filled. Only when the buffer is filled (or the input stream is finished)
    you can see the data visualized in visidata. That's why we use an adaptive buffer.
    For fast input streams, the buffer becomes large, for slow input streams, the buffer stays
    small"""
    buffer_size = 8
    processed_buffer_size = 0
    t_read = 0
    t_fill_target = 1   #in seconds
    while True:
        t_preread = time.time()
        next_chunk = fp.read(buffer_size)
        t_postread = time.time()
        if not next_chunk:
            break
        yield next_chunk
        t_read += t_postread - t_preread
        processed_buffer_size += len(next_chunk)

        speed_ratio = t_read / t_fill_target
        if speed_ratio <= 0.5:
            # if filling the buffer takes less than half the ideal time, double the size of the buffer.
            buffer_size = min(buffer_size * 2, max_buffer_size)
        else:
            # adjust the buffer size proportionately to how long it took to fill
            buffer_size = math.ceil(min(processed_buffer_size / speed_ratio, max_buffer_size))
            processed_buffer_size = 0
            t_read = 0

def splitter(stream, delim='\n'):
    'Generates one line/row/record at a time from stream, separated by delim'

    buf = type(delim)()

    for chunk in stream:
        buf += chunk

        *rows, buf = buf.split(delim)
        yield from rows

    buf = buf.rstrip(delim)  # trim empty trailing lines
    if buf:
        yield from buf.rstrip(delim).split(delim)


# rowdef: list
class TsvSheet(SequenceSheet):
    def iterload(self):
        delim = self.source.options.delimiter
        rowdelim = self.source.options.row_delimiter
        if delim == '':
            vd.warning("using '\\x00' as field delimiter")
            delim = '\x00'  #2272
            self.options.regex_skip = ''
        if rowdelim == '':
            vd.warning("using '\\x00' as row delimiter")
            rowdelim = '\x00'
            self.options.regex_skip = ''
        if delim == rowdelim:
            vd.fail('field delimiter and row delimiter cannot be the same')

        with self.open_text_source() as fp:
                regex_skip = getattr(fp, '_regex_skip', None)
                for line in splitter(adaptive_bufferer(fp), rowdelim):
                    if not line or (regex_skip and regex_skip.match(line)):
                        continue

                    row = line.split(delim)

                    if len(row) < self.nVisibleCols:
                        # extend rows that are missing entries
                        row.extend([None]*(self.nVisibleCols-len(row)))

                    yield row


@VisiData.api
def save_tsv(vd, p, vs):
    'Write sheet to file `fn` as TSV.'
    unitsep = p.options.delimiter
    rowsep = p.options.row_delimiter
    if unitsep == '':
        vd.warning("saving with '\\x00' as field delimiter")
        unitsep = '\x00'
    if rowsep == '':
        vd.warning("saving with '\\x00' as row delimiter")
        rowsep = '\x00'
    if unitsep == rowsep:
        vd.fail('field delimiter and row delimiter cannot be the same')
    trdict = vs.safe_trdict(delimiter=unitsep)

    with p.open(mode='w', encoding=vs.options.save_encoding) as fp:
        colhdr = unitsep.join(col.name.translate(trdict) for col in vs.visibleCols) + rowsep
        fp.write(colhdr)

        for dispvals in vs.iterdispvals(format=True, delimiter=unitsep):
            fp.write(unitsep.join(dispvals.values()))
            fp.write(rowsep)


vd.addGlobals({
    'TsvSheet': TsvSheet,
})



from copy import copy, deepcopy
import shutil
import subprocess
import io
import sys
import functools
import os
import itertools
import platform

from visidata import VisiData, vd, asyncthread, SettableColumn
from visidata import Sheet, Path, Column

if (
    # Windows
    sys.platform == 'win32'
    # WSL 2
    or "microsoft-standard-WSL2" in platform.uname().release
    # WSL 1
    or sys.platform == 'linux' and platform.uname().release.endswith("-Microsoft")
):
    syscopy_cmd_default = 'clip.exe'
    syspaste_cmd_default = 'powershell.exe -noprofile -command Get-Clipboard'
elif sys.platform == 'darwin':
    syscopy_cmd_default = 'pbcopy w'
    syspaste_cmd_default = 'pbpaste'
else:
    if 'WAYLAND_DISPLAY' in os.environ:
        syscopy_cmd_default = 'wl-copy'
        syspaste_cmd_default = 'wl-paste'
    else:
        syscopy_cmd_default = 'xclip -selection clipboard -filter'  # xsel --clipboard --input
        syspaste_cmd_default = 'xclip -selection clipboard -o'  # xsel --clipboard

vd.option('clipboard_copy_cmd', syscopy_cmd_default, 'command to copy stdin to system clipboard', sheettype=None)
vd.option('clipboard_paste_cmd', syspaste_cmd_default, 'command to send contents of system clipboard to stdout', sheettype=None)

@VisiData.api
def setClipboardRows(vd, rows):
    vd.memory.cliprows = rows

@VisiData.api
def getClipboardRows(vd):
    return vd.memory.cliprows

@VisiData.api
def setClipboardCols(vd, cols):
    vd.memory.clipcols = cols

@VisiData.api
def getClipboardCols(vd):
    return vd.memory.clipcols

@Sheet.api
def copyRows(sheet, rows):
    vd.setClipboardRows(rows)
    vd.setClipboardCols(list(sheet.visibleCols))
    if not rows:
        vd.warning('no %s selected; clipboard emptied' % sheet.rowtype)
    else:
        vd.status('copied %d %s to clipboard' % (len(rows), sheet.rowtype))

@Sheet.api
def copyCells(sheet, col, rows):
    vd.memory.clipcells = [col.getTypedValue(r) for r in rows]
    if not rows:
        vd.warning('no %s selected; clipboard emptied' % sheet.rowtype)
        return
    vd.status('copied %d %s.%s to clipboard' % (len(rows), sheet.rowtype, col.name))


@Sheet.api
def syscopyValue(sheet, val):
    # pipe val to stdin of clipboard command

    subprocess.run(
        sheet.options.clipboard_copy_cmd.split(),
        input=val,
        encoding='utf-8',
        stdout=subprocess.DEVNULL)

    vd.status('copied value to system clipboard')

@Sheet.api
def setColClipboard(sheet):
    if not vd.memory.clipcells:
        vd.warning("nothing to paste from clipcells")
        return
    sheet.cursorCol.setValuesTyped(sheet.onlySelectedRows, *vd.memory.clipcells)

@Sheet.api
def syscopyCells(sheet, cols, rows, filetype=None):
    filetype = filetype or vd.input("copy %d %s as filetype: " % (len(rows), sheet.rowtype), value=sheet.options.save_filetype or 'tsv')
    sheet.syscopyCells_async(cols, rows, filetype)


@Sheet.api
@asyncthread
def syscopyCells_async(sheet, cols, rows, filetype):
    vs = copy(sheet)
    vs.rows = rows or vd.fail('no %s selected' % sheet.rowtype)
    vs.columns = cols

    vd.status(f'copying {vs.nRows} {vs.rowtype} to system clipboard as {filetype}')

    with io.StringIO() as buf:
        with vd.TempFile() as temp:
            temp.close()  #2118

            vd.sync(vd.saveSheets(Path(f'{temp.name}.{filetype}', fptext=buf), vs, confirm_overwrite=False))
            subprocess.run(
                sheet.options.clipboard_copy_cmd.split(),
                input=buf.getvalue(),
                encoding='utf-8',
                stdout=subprocess.DEVNULL)


@VisiData.api
def sysclipValue(vd):
    return subprocess.check_output(vd.options.clipboard_paste_cmd.split()).decode('utf-8')


@VisiData.api
@asyncthread
def pasteFromClipboard(vd, cols, rows):
    text = vd.getLastArgs() or vd.sysclipValue().strip() or vd.fail('nothing to paste from system clipboard')

    vd.addUndoSetValues(cols, rows)
    lines = text.split('\n')
    if not lines:
        vd.warning('nothing to paste from system clipboard')
        return

    vs = cols[0].sheet
    newrows = [vs.newRow() for i in range(len(lines)-len(rows))]
    if newrows:
        rows.extend(newrows)
        vs.addRows(newrows)

    for line, r in zip(lines, rows):
        for v, c in zip(line.split('\t'), cols):
            c.setValue(r, v)


@Sheet.api
def delete_row(sheet, rowidx):
    if not sheet.rows:
        vd.fail("no row to delete")
    if not sheet.defer:
        oldrow = sheet.rows.pop(rowidx)
        vd.addUndo(sheet.rows.insert, rowidx, oldrow)
        # clear the deleted row from selected rows
        if sheet.isSelected(oldrow):
            sheet.addUndoSelection()
            sheet.unselectRow(oldrow)
        sheet.commitDeleteRow(oldrow)
    else:
        oldrow = sheet.rows[rowidx]
        sheet.rowDeleted(oldrow)

    sheet.setModified()
    return oldrow


@Sheet.api
@asyncthread
def paste_after(sheet, rowidx):
    'Paste rows from *vd.cliprows* at *rowidx*.'
    if not vd.getClipboardRows():  #1793
        vd.warning('nothing to paste from cliprows')
        return

    for col in vd.getClipboardCols()[sheet.nVisibleCols:]:
        newcol = SettableColumn()
        newcol.__setstate__(col.__getstate__())
        sheet.addColumn(newcol)

    addedRows = []

    for extrow in vd.getClipboardRows():
        if isinstance(extrow, Column):
            newrow = copy(extrow)
        else:
            newrow = sheet.newRow()
            for col, extcol in zip(sheet.visibleCols, vd.getClipboardCols()):
                col.setValue(newrow, extcol.getTypedValue(extrow))

        addedRows.append(newrow)

    sheet.addRows(addedRows, index=rowidx)


Sheet.addCommand('y', 'copy-row', 'copyRows([cursorRow])', 'copy current row to internal clipboard')

Sheet.addCommand('p', 'paste-after', 'paste_after(cursorRowIndex)', 'paste internal clipboard rows after current row')
Sheet.addCommand('P', 'paste-before', 'paste_after(cursorRowIndex-1)', 'paste internal clipboard rows before current row')

Sheet.addCommand('gy', 'copy-selected', 'copyRows(onlySelectedRows)', 'copy selected rows to internal clipboard')

Sheet.addCommand('zy', 'copy-cell', 'copyCells(cursorCol, [cursorRow]); vd.memoValue("clipval", cursorTypedValue, cursorDisplay)', 'copy current cell to internal clipboard')
Sheet.addCommand('zp', 'paste-cell', 'cursorCol.setValuesTyped([cursorRow], vd.memory.clipval) if vd.memory.clipval else vd.warning("nothing to paste")', 'set contents of current cell to last internal clipboard value')

Sheet.addCommand('d', 'delete-row', 'delete_row(cursorRowIndex); defer and cursorDown(1)', 'delete current row')
Sheet.addCommand('gd', 'delete-selected', 'deleteSelected()', 'delete selected rows')
Sheet.addCommand('zd', 'delete-cell', 'cursorCol.setValues([cursorRow], options.null_value)', 'delete current cell (set to None)')
Sheet.addCommand('gzd', 'delete-cells', 'cursorCol.setValues(onlySelectedRows, options.null_value)', 'delete contents of current column for selected rows (set to None)')

Sheet.bindkey('BUTTON2_PRESSED', 'go-mouse')
Sheet.addCommand('BUTTON2_RELEASED', 'syspaste-cells', 'pasteFromClipboard(visibleCols[cursorVisibleColIndex:], rows[cursorRowIndex:])', 'paste from system clipboard to region starting at cursor')
Sheet.bindkey('BUTTON2_CLICKED', 'go-mouse')
Sheet.bindkey('zP', 'syspaste-cells')
Sheet.addCommand('gzP', 'syspaste-cells-selected', 'pasteFromClipboard(visibleCols[cursorVisibleColIndex:], someSelectedRows)', 'paste from system clipboard to selected cells')

Sheet.addCommand('gzy', 'copy-cells', 'copyCells(cursorCol, onlySelectedRows)', 'copy contents of current column for selected rows to internal clipboard')
Sheet.addCommand('gzp', 'setcol-clipboard', 'setColClipboard()', 'set cells of current column for selected rows to last internal clipboard value')

Sheet.addCommand('Y', 'syscopy-row', 'syscopyCells(visibleCols, [cursorRow])', 'copy current row to system clipboard (using options.clipboard_copy_cmd)')

Sheet.addCommand('gY', 'syscopy-selected', 'syscopyCells(visibleCols, onlySelectedRows)', 'copy selected rows to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('zY', 'syscopy-cell', 'syscopyValue(cursorFullDisplay)', 'copy current cell to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('', 'syscopy-colname', 'syscopyValue(cursorCol.name)', 'copy current column header to system clipboard (using options.clipboard_copy_cmd)')
Sheet.addCommand('gzY', 'syscopy-cells', 'syscopyCells([cursorCol], onlySelectedRows, filetype="txt")', 'copy contents of current column from selected rows to system clipboard (using options.clipboard_copy_cmd')

Sheet.addCommand('x', 'cut-row', 'copyRows([sheet.delete_row(cursorRowIndex)]); defer and cursorDown(1)', 'delete current row and move it to internal clipboard')
Sheet.addCommand('gx', 'cut-selected', 'copyRows(onlySelectedRows); deleteSelected()', 'delete selected rows and move them to internal clipboard')
Sheet.addCommand('zx', 'cut-cell', 'copyCells(cursorCol, [cursorRow]); cursorCol.setValues([cursorRow], None)', 'delete current cell and move it to internal clipboard')
Sheet.addCommand('gzx', 'cut-cells', 'copyCells(cursorCol, onlySelectedRows); cursorCol.setValues(onlySelectedRows, None)', 'delete contents of current column for selected rows and move them to internal clipboard')


Sheet.bindkey('Del', 'delete-cell')
Sheet.bindkey('gDel', 'delete-cells')

vd.addMenuItems('''
    Edit > Delete > current row > delete-row
    Edit > Delete > current cell > delete-cell
    Edit > Delete > selected rows > delete-selected
    Edit > Delete > selected cells > delete-cells
    Edit > Copy > to internal clipboard > current cell > copy-cell
    Edit > Copy > to internal clipboard > current row > copy-row
    Edit > Copy > to internal clipboard > selected cells > copy-cells
    Edit > Copy > to internal clipboard > selected rows > copy-selected
    Edit > Copy > to system clipboard > current cell > syscopy-cell
    Edit > Copy > to system clipboard > current row > syscopy-row
    Edit > Copy > to system clipboard > selected cells > syscopy-cells
    Edit > Copy > to system clipboard > selected rows > syscopy-selected
    Edit > Cut > current row > cut-row
    Edit > Cut > selected cells > cut-selected
    Edit > Cut > current cell > cut-cell
    Edit > Paste > from internal clipboard > row after > paste-after
    Edit > Paste > from internal clipboard > row before > paste-before
    Edit > Paste > from internal clipboard > into selected cells > setcol-clipboard
    Edit > Paste > from internal clipboard > into current cell > paste-cell
    Edit > Paste > from system clipboard > cells at cursor > syspaste-cells
    Edit > Paste > from system clipboard > selected cells > syspaste-cells-selected
''')



import pytest

from visidata import Sheet, ColumnItem


@pytest.mark.usefixtures('curses_setup')
class TestColumnCache:
    def _sheet(self):
        class T(Sheet):
            columns = [ColumnItem('a', 'a')]
            def reload(self):
                self.rows = [{'a': 1}, {'a': 2}]
        s = T('t')
        s.reload()
        return s, s.column('a')

    def test_setValue_invalidates_cache(self):
        'setValue on a cached column must not leave a stale cached value.  #3155'
        s, c = self._sheet()
        c.resetCache()
        assert c.getValue(s.rows[0]) == 1  # populate cache
        c.setValue(s.rows[0], 99, setModified=False)
        assert c.getValue(s.rows[0]) == 99

    def test_undo_reverts_cached_cell(self):
        'Reverting a cell value (as undo does) shows immediately on a cached column.  #3155'
        s, c = self._sheet()
        c.resetCache()
        c.getValue(s.rows[0])                       # prime cache with original
        c.setValue(s.rows[0], 99, setModified=False)  # edit-cell
        assert c.getValue(s.rows[0]) == 99           # redraw repopulates cache
        c.setValue(s.rows[0], 1, setModified=False)  # undo's setValue
        assert c.getValue(s.rows[0]) == 1            # revert visible without another recalc



import stat

import pytest

from visidata import vd


# rowdef: n/a — exercises the $EDITOR launch+readback path that Ctrl+O uses
# during cell editing (visidata/_input.py: `edit_v = vd.launchExternalEditor(v)`)
# and that DirSheet sysopen-row uses. A non-interactive $EDITOR script stands in
# for the human, so the launch/readback is assertable headlessly (manual-tests.md #18).
class TestLaunchEditor:
    def _editor(self, tmp_path, body):
        'Write an executable $EDITOR script whose body acts on "$1" (the temp file).'
        p = tmp_path / 'fakeeditor.sh'
        p.write_text('#!/bin/sh\n' + body + '\n')
        p.chmod(p.stat().st_mode | stat.S_IXUSR)
        return str(p)

    def test_readback(self, tmp_path, monkeypatch):
        # editor rewrites the file; launchExternalEditor returns the new contents
        monkeypatch.setenv('EDITOR', self._editor(tmp_path, 'printf EDITED-"$(cat "$1")" > "$1"'))
        assert vd.launchExternalEditor('hello') == 'EDITED-hello'

    def test_trailing_newlines_trimmed(self, tmp_path, monkeypatch):
        # the function rstrips inevitable trailing newlines a real editor adds
        monkeypatch.setenv('EDITOR', self._editor(tmp_path, 'printf "one\\ntwo\\n\\n" > "$1"'))
        assert vd.launchExternalEditor('x') == 'one\ntwo'

    def test_unchanged_roundtrips(self, tmp_path, monkeypatch):
        # editor that leaves the file alone returns the original value
        monkeypatch.setenv('EDITOR', self._editor(tmp_path, ':'))
        assert vd.launchExternalEditor('keepme') == 'keepme'

    def test_no_editor_fails(self, monkeypatch):
        monkeypatch.delenv('EDITOR', raising=False)
        monkeypatch.delenv('VISUAL', raising=False)
        with pytest.raises(Exception):
            vd.launchExternalEditor('hello')



from visidata import vd, VisiData, TsvSheet


@VisiData.api
def open_usv(vd, p):
    p.options.set('delimiter', '\u241f', p, cmdlog=False)
    p.options.set('row_delimiter', '\u241e', p, cmdlog=False)
    return UsvSheet(p.base_stem, source=p)


class UsvSheet(TsvSheet):
    pass


@VisiData.api
def save_usv(vd, p, vs):
    p.options.set('delimiter', '\u241f', p, cmdlog=False)
    p.options.set('row_delimiter', '\u241e', p, cmdlog=False)
    vd.save_tsv(p, vs)



import visidata
from visidata.loaders.csv import CsvSheet
from visidata.loaders.tsv import TsvSheet


def _load(tmp_path, name, text, cls):
    p = tmp_path / name
    p.write_text(text)
    vs = cls(name, source=visidata.Path(str(p)))
    vs.reload.__wrapped__(vs)
    return vs


def test_csv_skips_blank_lines(tmp_path):
    # #3085: the csv loader must drop blank lines like the tsv loader,
    # rather than keeping a phantom empty row.
    data = 'col\n1\n\n3\n'
    csvs = _load(tmp_path, 'blank.csv', data, CsvSheet)
    tsvs = _load(tmp_path, 'blank.tsv', data, TsvSheet)
    assert len(csvs.rows) == len(tsvs.rows) == 2  # '1' and '3'; blank line gone


def test_csv_keeps_row_of_empty_values(tmp_path):
    # a ',,' line is a real row whose values happen to be empty (not a blank
    # line), so it must be preserved.
    vs = _load(tmp_path, 'empties.csv', 'col,a,b\n1,2,3\n,,\n7,8,9\n', CsvSheet)
    assert len(vs.rows) == 3



from visidata import vd, BaseSheet, Column
from visidata.pyobj import PythonSheet, PyobjSheet


# rowdef: Distribution object from importlib.metadata
class PythonPackagesSheet(PythonSheet):
    'Sheet displaying installed Python packages via importlib.metadata.distributions()'
    rowtype = 'packages'
    columns = [
        Column('name', getter=lambda c,r: r.name),
        Column('version', getter=lambda c,r: r.version),
        Column('summary', getter=lambda c,r: r.metadata.get('Summary', '')),
        Column('author', getter=lambda c,r: r.metadata.get('Author', '')),
        Column('license', getter=lambda c,r: r.metadata.get('License', '')),
        Column('location', getter=lambda c,r: str(r._path.parent) if hasattr(r, '_path') and r._path else ''),
        Column('home_page', getter=lambda c,r: r.metadata.get('Home-page', '')),
    ]
    nKeys = 1

    def reload(self):
        import importlib.metadata
        distributions = list(importlib.metadata.distributions())
        try:
            self.rows = sorted(distributions, key=lambda d: d.name.lower())
        except AttributeError:  # Python == 3.9       distributions do not have .name attr, instead use .metadata.distributions['Name']
            for r in distributions:
                r.name = r.metadata['Name']
            self.rows = sorted(distributions, key=lambda d: d.name.lower())

    def openRow(self, row):
        'Open package metadata as Python object'
        return PyobjSheet(f'{row.name}-{row.version}', source=row)


BaseSheet.addCommand('', 'open-python-packages', 'vd.push(PythonPackagesSheet("python-packages"))', 'open Python Packages Sheet listing the installed packages')

vd.addMenuItems('''
    System > Python > installed packages > open-python-packages
''')

vd.addGlobals(PythonPackagesSheet=PythonPackagesSheet)



'''Conditional formatting commands.  #3061'''

from visidata import vd, Sheet, CellColorizer, RowColorizer, ColumnColorizer


# default colors offered when adding a conditional format
condfmt_default_colors = [
    'red', 'green', 'yellow', 'cyan', 'magenta', 'blue',
    'on red', 'on green', 'on yellow', 'on cyan', 'on magenta', 'on blue',
]

condfmt_precedence = 8


vd.help_condfmt_expr = '''# Conditional Format Expression

A Python expression evaluated against each cell.  Color is applied where the expression is truthy.

## Bindings

- bare column names resolve to that column's typed value for the current row (e.g. `temperature > 100` on a sheet with a `temperature` column)
- `value` / `v` -- the typed value of the current cell
- `sheet` / `s`, `col` / `c`, `row` / `r` -- the sheet, column, and row

## Examples

    value < 0
    score >= 90
    status == 'failed'
    'error' in value
    c.name.startswith('total_')

## Notes

- Row colorizers fire per cell, so `value` refers to the current cell, not "the row's value".  For row-level rules, reference a sibling column by name (e.g. `status == 'failed'`).
- Cells where the expression raises are silently not colored.  Run vd with `--debug` to see exceptions in the status log.
'''


vd.help_condfmt_color = '''# Conditional Format Color

A color string or a `color_*` option name.

## Color syntax

    <attribute> <fg-color> on <bg-color>

- attributes: [:bold]bold[/] [:underline]underline[/] [:italic]italic[/] [:reverse]reverse[/]
- colors: 0-255 or [:black on 238]black[/] [:red on 238]red[/] [:green on 238]green[/] [:yellow on 238]yellow[/] [:blue on 238]blue[/] [:magenta on 238]magenta[/] [:cyan on 238]cyan[/] [:white on 238]white[/]
- a second color is used as a fallback if the first is not available

## Color option name

Any `color_*` option name works as a synonym, with or without the `color_` prefix -- the conditional format picks up whatever the active theme defines:

- `error` -> `color_error`
- `current_row`, `selected_row`, `key_col`, ...

## Examples

    red
    bold green
    on red
    215 yellow on 17
    error

See [:onclick https://visidata.org/docs/colors]https://visidata.org/docs/colors[/] for more detail.
'''


class CondFmtContext(dict):
    '''Eval-context for conditional formatting expressions.

    Exposes the four colorizer arguments under both short (s/c/r/v) and
    long (sheet/col/row/value) names, and falls back to bare column-name
    lookups for the current row -- like `addcol-expr` and friends.

    Note: this duplicates a bit of LazyComputeRow's bare-column-name
    resolution rather than calling Sheet.evalExpr, because LCR's `_lcm`
    is cached on col/sheet and would pin the first LCR's `extra` AttrDict
    across calls -- so cell-specific bindings (`value`, `v`) would leak
    between cells.  See sheets.py:91-95.
    '''
    def __init__(self, sheet, col, row, value):
        super().__init__(s=sheet, sheet=sheet,
                         c=col, col=col,
                         r=row, row=row,
                         v=value, value=value)
        self._sheet = sheet
        self._row = row

    def __missing__(self, k):
        col = self._sheet.colsByName.get(k)
        if col is not None:
            return col.getTypedValue(self._row)
        raise KeyError(k)


@Sheet.api
def pick_unused_color(sheet):
    'Return the first default conditional-format color not already used by a colorizer on this sheet.'
    used = {cz.coloropt for cz in sheet._colorizers if cz.coloropt}
    for c in condfmt_default_colors:
        if c not in used:
            return c
    return condfmt_default_colors[0]


def _make_color_func(expr):
    'Compile *expr* once and return a Colorizer func evaluated against a CondFmtContext.'
    code = compile(expr, '<color expr>', 'eval')
    def func(s, c, r, v):
        # _colorize passes a DisplayWrapper as v; expressions want the typed value
        typedval = v.typedval if hasattr(v, 'typedval') else v
        try:
            return eval(code, vd.getGlobals(), CondFmtContext(s, c, r, typedval))
        except Exception as e:  # cells where the expr cannot evaluate are simply not colored
            vd.debug(f'condfmt `{expr}`: {type(e).__name__}: {e}')
            return False
    return func


def _add_condfmt(sheet, ColorizerClass, noun, color, expr):
    sheet.addColorizer(ColorizerClass(condfmt_precedence, color, _make_color_func(expr)))
    vd.status(f'{noun} where `{expr}` colored `{color}`')


@Sheet.api
def color_cell(sheet, color, expr):
    'Add a CellColorizer that paints cells with *color* where *expr* is truthy.'
    _add_condfmt(sheet, CellColorizer, 'cells', color, expr)


@Sheet.api
def color_row(sheet, color, expr):
    'Add a RowColorizer that paints rows with *color* where *expr* is truthy.'
    _add_condfmt(sheet, RowColorizer, 'rows', color, expr)


@Sheet.api
def color_col(sheet, color, expr):
    'Add a ColumnColorizer that paints columns with *color* where *expr* is truthy.'
    _add_condfmt(sheet, ColumnColorizer, 'columns', color, expr)


@Sheet.api
def inputCondFmt(sheet):
    'Prompt for color and condition expression for conditional formatting.'
    return vd.inputMultiple(
        expr=dict(prompt='expr: ',  type='expr',  help=vd.help_condfmt_expr),
        color=dict(prompt='color: ', value=sheet.pick_unused_color(), help=vd.help_condfmt_color),
    )


Sheet.addCommand('', 'color-cell', 'color_cell(**inputCondFmt())', 'apply color to cells where Python expr is truthy')
Sheet.addCommand('', 'color-row',  'color_row(**inputCondFmt())',  'apply color to rows where Python expr is truthy')
Sheet.addCommand('', 'color-col',  'color_col(**inputCondFmt())',  'apply color to columns where Python expr is truthy')

vd.addMenuItems('''
    View > Color > cells where > color-cell
    View > Color > rows where > color-row
    View > Color > columns where > color-col
''')


## tests

def _condfmt_test_sheet():
    'Build a tiny TableSheet with typed columns for condfmt tests.'
    from visidata import Sheet, Column
    return Sheet('toy', columns=[
        Column('temperature', type=int, getter=lambda c,r: r['temperature']),
        Column('status',      type=str, getter=lambda c,r: r['status']),
    ], rows=[
        {'temperature': 50,  'status': 'ok'},
        {'temperature': 120, 'status': 'failed'},
        {'temperature': -5,  'status': 'cold'},
    ])


def _condfmt_added(sheet):
    'Return colorizers added by condfmt commands (filters by closure qualname).'
    return [c for c in sheet._colorizers
            if getattr(c._func, '__qualname__', '') == '_make_color_func.<locals>.func']


def test_condfmt_bare_column_name(vd):  #3061
    'Bare column names in expressions resolve to the typed value of that column.'
    from visidata import CellColorizer
    s = _condfmt_test_sheet()
    s.color_cell('red', 'temperature > 100')
    cz, = _condfmt_added(s)
    assert isinstance(cz, CellColorizer)
    assert cz.coloropt == 'red'
    assert cz.func(s, s.columns[0], s.rows[0], 50)  is False  # 50 not > 100
    assert cz.func(s, s.columns[0], s.rows[1], 120) is True   # 120 > 100
    assert cz.func(s, s.columns[0], s.rows[2], -5)  is False  # -5 not > 100


def test_condfmt_value_binding(vd):  #3061
    "`value` and `v` resolve to the current cell's typed value."
    s = _condfmt_test_sheet()
    s.color_cell('red', 'value < 0')
    cz, = _condfmt_added(s)
    assert cz.func(s, s.columns[0], s.rows[2], -5) is True
    assert cz.func(s, s.columns[0], s.rows[0], 50) is False

    s = _condfmt_test_sheet()
    s.color_cell('red', 'v < 0')
    cz, = _condfmt_added(s)
    assert cz.func(s, s.columns[0], s.rows[2], -5) is True
    assert cz.func(s, s.columns[0], s.rows[0], 50) is False


def test_condfmt_col_binding(vd):  #3061
    "`c` and `col` resolve to the column being queried, supporting attr access."
    s = _condfmt_test_sheet()
    s.color_col('underline', 'c.name.startswith("temp")')
    cz, = _condfmt_added(s)
    assert cz.func(s, s.columns[0], s.rows[0], 50)    is True   # temperature col
    assert cz.func(s, s.columns[1], s.rows[0], 'ok')  is False  # status col


def test_condfmt_row_binding_string_column(vd):  #3061
    'Row colorizer expression referencing a sibling column by bare name.'
    from visidata import RowColorizer
    s = _condfmt_test_sheet()
    s.color_row('on red', "status == 'failed'")
    cz, = _condfmt_added(s)
    assert isinstance(cz, RowColorizer)
    # row colorizer fires per-cell but truth value is based on the row's status col
    assert cz.func(s, s.columns[0], s.rows[1], 120) is True   # status='failed' row
    assert cz.func(s, s.columns[1], s.rows[1], 'failed') is True
    assert cz.func(s, s.columns[0], s.rows[0], 50)  is False  # status='ok' row


def test_condfmt_silent_eval_error(vd):  #3061
    'Expression errors return False so unrenderable cells are simply not colored.'
    s = _condfmt_test_sheet()
    s.color_cell('red', 'undefined_name > 0')
    cz, = _condfmt_added(s)
    assert cz.func(s, s.columns[0], s.rows[0], 50) is False


def test_pick_unused_color(vd):  #3061
    'pick_unused_color skips colors already in use by colorizers on this sheet.'
    s = _condfmt_test_sheet()
    assert s.pick_unused_color() == 'red'
    s.color_cell('red', '0')
    assert s.pick_unused_color() == 'green'
    s.color_row('green', '0')
    assert s.pick_unused_color() == 'yellow'


def test_condfmt_color_option_name(vd):  #3061
    'A color option name like `error` resolves to that option (color_error).'
    from visidata import colors
    s = _condfmt_test_sheet()
    s.color_cell('error', 'value < 0')  # `error` should resolve to color_error
    cz, = _condfmt_added(s)
    assert cz.coloropt == 'error'  # stored verbatim; resolution happens at draw time
    # colors.get_color('error') falls back to color_error option
    assert colors.get_color('error').colorname == colors.get_color('color_error').colorname


def test_condfmt_value_is_typed_not_wrapper(vd):  #3061
    "`value`/`v` are the typed cell value, not a DisplayWrapper -- so `'a' in value` works."
    s = _condfmt_test_sheet()
    s.color_cell('red', "'a' in value")
    cz, = _condfmt_added(s)
    # invoke through the same call path as draw: _colorize passes a DisplayWrapper
    cellval = s.columns[1].getCell(s.rows[1])  # status='failed' -> contains 'a'
    assert cz.func(s, s.columns[1], s.rows[1], cellval) is True
    cellval = s.columns[1].getCell(s.rows[0])  # status='ok' -> no 'a'
    assert cz.func(s, s.columns[1], s.rows[0], cellval) is False



__all__ = ['__title__', '__author__', '__version__', '__description__', '__license__', '__copyright__']

__title__ = 'vdsql'
__author__ = 'Saul Pwanson <code@saul.pw>'
__version__ = '0.3dev'
__description__ = 'VisiData for database queries'
__license__ = 'Apache License, Version 2.0'
__copyright__ = 'Copyright 2022 ' + __author__



__author__ = 'Andy Craig, andycraig (https://github.com/andycraig)'


import re

from visidata import VisiData, vd, Sheet, Column, floatsi, currency, date

FLOAT_FORMAT_RE = r'\{:\.([0-9]+)f\}|%\.([0-9]+)f'

date_fmtstrs = [
    '%Y',
    '%Y-%m',
#    '%Y-W%U',
    '%Y-%m-%d',
    '%Y-%m-%d %H',
    '%Y-%m-%d %H:%M',
    '%Y-%m-%d %H:%M:%S',
    '%Y-%m-%d %H:%M:%S.%f',
]

@Column.api
def setcol_precision(col, amount:int):
    if col.type is date:
        try:
            i = date_fmtstrs.index(col.fmtstr)
        except ValueError:
            i = 2
        col.fmtstr = date_fmtstrs[(i+amount)%len(date_fmtstrs)]
    elif col.type in (float, floatsi, currency):
        if col.fmtstr == '':
            col.fmtstr = f'%.{2 + amount}f'
        else:
            m = re.fullmatch(FLOAT_FORMAT_RE, col.fmtstr)
            if not m:
                vd.fail('could not parse column fmtstr')
            if m[1]:
                col.fmtstr = '{:.' + f'{max(0, int(m[1]) + amount)}f' + '}'
            elif m[2]:
                col.fmtstr =       f'%.{max(0, int(m[2]) + amount)}f'
    else:
        vd.fail('column type must be numeric or date')

@VisiData.api
def setopt_precision(vd, precision):
     m = re.fullmatch(FLOAT_FORMAT_RE, vd.options.disp_float_fmt)
     if not m:
         vd.fail('could not parse disp_float_fmt')
     if m[1]:
         vd.options.disp_float_fmt = '{:.' + str(precision) + 'f}'
     elif m[2]:
         vd.options.disp_float_fmt = '%.0' + str(precision) + 'f'

vd.addMenuItems('''
    Column > Set precision > more > setcol-precision-more
    Column > Set precision > less > setcol-precision-less
''')

Sheet.addCommand('Alt+-', 'setcol-precision-less', 'cursorCol.setcol_precision(-1)', 'show less precision in current column')
Sheet.addCommand('Alt++', 'setcol-precision-more', 'cursorCol.setcol_precision(1)', 'show more precision in current column')
Sheet.addCommand('g%', 'setcol-precision-input', 'precision = max(0, int(vd.input("float precision=", value="2"))); vd.setopt_precision(precision)', 'change the number of decimal places shown for floats')



from visidata import vd


# v3.4 prettykeys overhaul (#2594): curses key codes render as human-readable
# names. Pure translation, asserted directly.
class TestPrettyKeys:
    def test_prettykeys(self):
        assert vd.prettykeys('^S') == 'Ctrl+S'
        assert vd.prettykeys('^M') == 'Enter'
        assert vd.prettykeys(' ') == 'Space'
        assert vd.prettykeys('kUP') == 'Shift+Up'      # curses shifted-arrow code
        assert vd.prettykeys('KEY_HOME') == 'Home'
        assert vd.prettykeys('g^S') == 'gCtrl+S'        # g prefix preserved, ^ translated



from visidata import vd


def test_parsePos():
    'Test cases from PR #2425'
    inputs = [('foo.csv', {})]

    assert vd.parsePos('') is None
    assert vd.parsePos('2', inputs=inputs) == ([0], None, '2')  # +row with inputs: last sheet (index 0)
    assert vd.parsePos('2', inputs=None) == (None, None, '2')  # +row without inputs: current sheet
    assert vd.parsePos('-1', inputs=inputs) == ([0], None, '-1')  # negative row index stays as string
    assert vd.parsePos('1:', inputs=inputs) == ([0], 1, None)  # +col:
    assert vd.parsePos(':2', inputs=inputs) == ([0], None, 2)  # +:row
    assert vd.parsePos('1:2', inputs=inputs) == ([0], 1, 2)  # +col:row
    assert vd.parsePos('name:value', inputs=inputs) == ([0], 'name', 'value')  # string col:row
    assert vd.parsePos(':1:', inputs=inputs) == ([], 1, None)  # +:col: all sheets
    assert vd.parsePos(':1:2') == ([], 1, 2)  # +:col:row all sheets
    assert vd.parsePos('::2') == ([], None, 2)  # +::row all sheets
    assert vd.parsePos('0:1:2') == ([0], 1, 2)  # +sheet:col:row
    assert vd.parsePos('1::') == ([1], None, None)  # +sheet:: switch sheet only
    assert vd.parsePos('-2::') == ([-2], None, None)  # negative sheet index
    assert vd.parsePos('0:a:1:2') == ([0, 'a'], 1, 2)  # named subsheet
    assert vd.parsePos('0:0:1:2') == ([0, 0], 1, 2)  # numeric subsheet index
    assert vd.parsePos('0:a:b:1:2') == ([0, 'a', 'b'], 1, 2)  # nested subsheets
    assert vd.parsePos(':a:1:2') == (['', 'a'], 1, 2)  # all sheets + subsheet



import os
import re
import shlex
import shutil
import stat
import subprocess
import contextlib
try:
    import pwd
    import grp
except ImportError:
    pwd = None # pwd,grp modules not available on Windows
    grp = None

from visidata import Column, Sheet, LazyComputeRow, asynccache, BaseSheet, vd
from visidata import Path, asyncthread, VisiData
from visidata import modtime, filesize, vstat, Progress, TextSheet
from visidata.type_date import date


vd.option('dir_depth', 0, 'folder recursion depth on DirSheet')
vd.option('dir_hidden', False, 'load hidden files on DirSheet')
vd.option('active_procs', 10, 'number of concurrent processes on DirSheet')

vd.spawnedProcesses = []

def bytes_rstrip(*args, **kwargs): #3081
    return bytes.rstrip(*args, **kwargs)


@VisiData.api
def popen(vd, *args, **kwargs):
    p = subprocess.Popen(*args, **kwargs)
    vd.spawnedProcesses.append(p)
    return p


@VisiData.api
def killLeftoverProcesses(vd):
    for p in vd.spawnedProcesses:
        if p.returncode is None:
            p.kill()


@VisiData.api
def guess_dir(vd, p):
    if p.is_dir():
        return dict(filetype='dir', _likelihood=10)


@VisiData.lazy_property
def currentDirSheet(p):
    'Support opening the current DirSheet from the vdmenu'
    return DirSheet(Path('.').absolute().name, source=Path('.'))

@asyncthread
def exec_shell(*args):
    p = vd.popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = p.communicate()
    if err or out:
        lines = err.decode('utf8').splitlines() + out.decode('utf8').splitlines()
        vd.push(TextSheet(' '.join(args), source=lines))


@VisiData.api
def open_dir(vd, p):
    if p.is_dir():
        return DirSheet(p.base_stem, source=p)
    if p.is_file():
        vd.status(f'opening {p.given} as txt')
        return vd.open_txt(p)
    vd.warning(f'could not determine file type for {p.given}')
    return DirSheet(p.base_stem, source=p)

@VisiData.api
def open_fdir(vd, p):
    return FileListSheet(p.base_stem, source=p)

@VisiData.api
def addShellColumns(vd, cmd, sheet, curcol=None):
    shellcol = ColumnShell(cmd, source=sheet, width=0, curcol=curcol)
    sheet.addColumnAtCursor(
            Column(cmd+'_stdout', type=bytes_rstrip, srccol=shellcol, getter=lambda col,row: col.srccol.getValue(row)[0]),
            Column(cmd+'_stderr', type=bytes_rstrip, srccol=shellcol, getter=lambda col,row: col.srccol.getValue(row)[1]),
            shellcol)


SHELL_COLREF_RE = r'\$\{([^}]+)\}|\$(\w+)'


class ColumnShell(Column):
    def __init__(self, name, cmd=None, curcol=None, **kwargs):
        super().__init__(name, **kwargs)
        self.expr = cmd or name
        self.curcol = curcol

    @asynccache(lambda col,row: (col, col.sheet.rowid(row)))
    def calcValue(self, row):
        try:
            context = LazyComputeRow(self.source, row, curcol=self.curcol)
            cmd = re.sub(SHELL_COLREF_RE,
                         lambda m: shlex.quote(str(context[m.group(1) or m.group(2)])),
                         self.expr)
            p = vd.popen([os.getenv('SHELL', 'bash'), '-c', cmd],
                    stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            return p.communicate()
        except Exception as e:
            vd.exceptionCaught(e)


class DirSheet(Sheet):
    'Sheet displaying directory, using Enter to open a particular file.  Edited fields are applied to the filesystem.'
    guide = '''
        # Directory Sheet
        This is a list of files in the {sheet.displaySource} folder.

        - {help.commands.open_row_file}
        - {help.commands.open_rows}
        - {help.commands.open_dir_parent}
        - {help.commands.sysopen_row}
        - {help.commands.open_preview}

        ## Options (must reload to take effect)

        - {help.options.dir_depth}
        - {help.options.dir_hidden}
    '''
    rowtype = 'files' # rowdef: Path
    defer = True
    columns = [
        Column('directory',
            getter=lambda col,row: str(row.parent) if str(row.parent) in ('.', '/') else str(row.parent) + '/',
            setter=lambda col,row,val: col.sheet.moveFile(row, val)),
        Column('filename',
            getter=lambda col,row: row.name,
            setter=lambda col,row,val: col.sheet.renameFile(row, val)),
        Column('abspath', width=0, type=str,
            getter=lambda col,row: row,
            setter=lambda col,row,val: os.rename(row, val)),
        Column('ext', getter=lambda col,row: row.is_dir() and '/' or row.ext),
        Column('size', type=int,
            getter=lambda col,row: filesize(row),
            setter=lambda col,row,val: os.truncate(row, int(val))),
        Column('modtime', type=date,
            getter=lambda col,row: modtime(row),
            setter=lambda col,row,val: os.utime(row, times=((row.stat().st_atime, float(val))))),
        Column('owner', width=0,
            getter=lambda col,row: pwd.getpwuid(row.stat().st_uid).pw_name,
            setter=lambda col,row,val: os.chown(row, pwd.getpwnam(val).pw_uid, -1)),
        Column('group', width=0,
            getter=lambda col,row: grp.getgrgid(row.stat().st_gid).gr_name,
            setter=lambda col,row,val: os.chown(row, -1, grp.getgrnam(val).pw_gid)),
        Column('mode', width=0,
            getter=lambda col,row: '{:o}'.format(row.stat().st_mode),
            setter=lambda col,row,val: os.chmod(row, int(val, 8))),
        Column('filetype', width=0, cache='async', getter=lambda col,row: vd.popen(['file', '--brief', row], stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()[0].strip()),
        Column('preview', width=0, cache=True, getter=lambda col,row: col.sheet._openPreview(row)),
    ]
    nKeys = 2
    _ordering = [('modtime', True), ('filename', False)]  # sort by reverse modtime initially

    @staticmethod
    def colorOwner(sheet, col, row, val):
        ret = ''
        if col.name == 'group':
            mode = row.stat().st_mode
            if mode & stat.S_IXGRP: ret = 'bold '
            if mode & stat.S_IWGRP: return ret + 'green'
            if mode & stat.S_IRGRP: return ret + 'yellow'
        elif col.name == 'owner':
            mode = row.stat().st_mode
            if mode & stat.S_IXUSR: ret = 'bold '
            if mode & stat.S_IWUSR: return ret + 'green'
            if mode & stat.S_IRUSR: return ret + 'yellow'

    def moveFile(self, row, newparent):
        parent = Path(newparent)
        newpath = Path(parent/row.name)
        if parent.exists():
            if not parent.is_dir():
                vd.error('destination %s not a directory' % parent)
        else:
            with contextlib.suppress(FileExistsError):
                os.makedirs(parent)

        row.rename(newpath)
        row.given = newpath # modify visidata.Path
        self.restat()

    def renameFile(self, row, val):
        newpath = row.with_name(val)
        row.rename(newpath)
        row.given = newpath
        self.restat()

    def removeFile(self, path):
        if path.is_dir():
            if self.options.safety_first:
                os.rmdir(path)
            else:
                shutil.rmtree(path)  #1965
        else:
            path.unlink()

    def commitDeleteRow(self, r):
        self.removeFile(r)

    def newRow(self):
        vd.fail('new file not supported')

    def iterload(self):
        hidden_files = self.options.dir_hidden

        def _walkfiles(p, dir_depth:int=0):
            basepath = str(p)
            for folder, subdirs, files in os.walk(basepath):
                subfolder = folder[len(basepath)+1:]
                if not hidden_files and subfolder.startswith('.'): continue
                if subfolder in ['.', '..']: continue

                fpath = Path(folder)
                if str(fpath) != str(p):
                    yield fpath

                for fn in files:
                    yield fpath/fn

                if dir_depth < len(fpath.parents)-len(p.parents)+1:
                    for d in subdirs:
                        yield fpath/d
                    subdirs.clear()

        for p in _walkfiles(self.source, self.options.dir_depth):
            if not hidden_files and p.name.startswith('.'):
                continue

            yield p

    def preloadHook(self):
        super().preloadHook()
        Path.stat.cache_clear()

    def restat(self):
        vstat.cache_clear()

    @asyncthread
    def putChanges(self):
        self.commitAdds()
        self.commitMods()
        self.commitDeletes()

        self._deferredDels.clear()
        self.reload()

    def getDefaultSaveName(sheet):
        return sheet.name + '.' + sheet.options.save_filetype


class FileListSheet(DirSheet):
    _ordering = []
    def iterload(self):
        for fn in self.source.open():
            yield Path(fn.rstrip())


@VisiData.api
def inputShell(vd):
    cmd = vd.input("sh$ ", type="sh")
    refs = [m.group(1) or m.group(2) for m in re.finditer(SHELL_COLREF_RE, cmd)]
    if not refs:
        vd.warning('no $column in command')
    else:
        colnames = [col.name for col in vd.sheet.columns]
        badnames = [name for name in refs if name not in colnames]
        if badnames:
            vd.fail(f'no such columns: {", ".join(badnames)}')
    return cmd

DirSheet.addCommand('`', 'open-dir-parent', 'vd.push(openSource(source.parent if source.resolve()!=Path(".").resolve() else os.path.dirname(source.resolve())))', 'open parent directory')  #1801
BaseSheet.addCommand('', 'open-dir-current', 'vd.push(vd.currentDirSheet)', 'open Directory Sheet: browse properties of files in current directory')

Sheet.addCommand('z;', 'addcol-shell', 'cmd=inputShell(); addShellColumns(cmd, sheet, curcol=cursorCol)', 'create new column from bash expression, with $columnNames as variables')

DirSheet.addCommand('Enter', 'open-row-file', 'vd.push(openSource(cursorRow or fail("no row"), filetype="dir" if cursorRow.is_dir() else LazyComputeRow(sheet, cursorRow).ext))', 'open current file as a new sheet')
DirSheet.addCommand('gEnter', 'open-rows', 'for r in selectedRows: vd.push(openSource(r))', 'open selected files as new sheets')
DirSheet.addCommand('Ctrl+O', 'sysopen-row', 'launchEditor(cursorRow)', 'open current file in external $EDITOR')
DirSheet.addCommand('gCtrl+O', 'sysopen-rows', 'launchEditor(*selectedRows)', 'open selected files in external $EDITOR')

DirSheet.addCommand('y', 'copy-row', 'copy_files([cursorRow], inputPath("copy to dest: "))', 'copy file to given directory *path*')
DirSheet.addCommand('gy', 'copy-selected', 'copy_files(selectedRows, inputPath("copy to dest: ", value=cursorRow.given))', 'copy selected files to given directory *path*')

DirSheet.addCommand('zEnter', 'open-row-filetype', 'ft = input("filetype: ", type="filetype", value=options.filetype or LazyComputeRow(sheet, cursorRow).ext); vd.push(openSource(cursorRow, filetype=ft) or fail(f"file {cursorDisplay} does not exist"))', 'open file in current row as input filetype')
DirSheet.addCommand('', 'open-preview', 'sheet.previewFile(cursorRow)', 'open split preview of file at cursor')


@DirSheet.api
def _openPreview(sheet, p):
    vs = vd.openSource(p, filetype="dir" if p.is_dir() else LazyComputeRow(sheet, p).ext)
    vs._dirpreview = True
    vs.ensureLoaded()
    return vs


@DirSheet.api
def previewFile(sheet, p):
    if not p: return
    vs = sheet.column('preview').getValue(p)
    if not isinstance(vs, BaseSheet):
        return  # still loading or error
    # remove old preview from pane 2
    for old in vd.sheetstack(2):
        if getattr(old, '_dirpreview', False):
            vd.sheets.remove(old)
    vd.push(vs, pane=2)
    vd.options.disp_splitwin_pct = vd.options.disp_splitwin_pct or 50
    sheet._previewing = True


@DirSheet.after
def checkCursor(sheet):
    if not getattr(sheet, '_previewing', False):
        return
    if not vd.options.disp_splitwin_pct:
        sheet._previewing = False
        return
    p = sheet.cursorRow
    if p and p != getattr(sheet, '_preview_path', None):
        sheet._preview_path = p
        sheet.previewFile(p)
    # preload previews for visible rows
    previewCol = sheet.column('preview')
    for row in sheet.rows[sheet.topRowIndex:sheet.topRowIndex + sheet.nScreenRows]:
        previewCol.getValue(row)


@DirSheet.api
@asyncthread
def copy_files(sheet, paths, dest):
    destdir = Path(dest)
    destdir.is_dir() or vd.fail('target must be directory')
    vd.status('copying %s %s to %s' % (len(paths), sheet.rowtype, destdir))
    os.makedirs(destdir, exist_ok=True)
    for srcpath in Progress(paths, gerund='copying'):
        try:
            destpath = destdir/str(srcpath._path.name)
            if srcpath.is_dir():
                shutil.copytree(srcpath, destpath)
            else:
                shutil.copyfile(srcpath, destpath)
        except Exception as e:
            vd.exceptionCaught(e)


vd.addGlobals(
    DirSheet=DirSheet,
    bytes_rstrip=bytes.rstrip,
)

vd.addMenuItems('''
    Column > Add column > shell > addcol-shell
    Row > Open file > open-row-filetype
    View > Preview file > open-preview
''')



'''History palette: shows previous inputs during search/regex/etc input.'''

from visidata import vd, VisiData, clipdraw, colors


class HistoryPalette:
    'Manages state, navigation, and drawing for the history input palette.'
    def __init__(self, sheet, items, input_height, updater):
        'Initialize palette state from history items.'
        self.sheet = sheet
        self.items = items                                    # raw history list
        self.haystack = [dict(input=item) for item in items]  # precomputed for fuzzymatch
        self.display = [(item, item) for item in items]       # (formatted_text, raw_value)
        self.cursor = -1       # -1 = free-typing, 0+ = index into display
        self.top = 0           # first visible row
        self.orig = ''         # saved input before history navigation
        self.input_height = input_height
        self.updater = updater # caller's original updater to chain

    def nvis(self):
        'Max visible rows for the palette given current window size.'
        return max(1, min(self.sheet.windowHeight - 2 - self.input_height, vd.options.disp_cmdpal_max))

    def bounds(self):
        'Clamp cursor and top, then enforce cursor visibility.'
        n = len(self.display)
        if n == 0: return
        nv = self.nvis()
        self.cursor = max(0, min(self.cursor, n - 1))
        self.top = max(0, min(self.top, n - 1))
        if self.top > self.cursor:              self.top = self.cursor
        if self.top + nv - 1 < self.cursor:     self.top = self.cursor - nv + 1

    def _recompute(self, value):
        'Recompute display list from fuzzy match against current input.'
        if value:
            matches = vd.fuzzymatch(self.haystack, value.split())
            self.display = [(m.formatted.get('input', m.match['input']), m.match['input']) for m in matches]
        else:
            self.display = [(item, item) for item in self.items]

    def enter_nav(self, v):
        'Save orig and place cursor at end of display list.'
        self.orig = v
        self._recompute(v)
        self.cursor = len(self.display) - 1
        self.top = max(0, self.cursor - self.nvis() + 1)

    def sel(self):
        'Return (value, cursor_pos) for current cursor selection.'
        v = self.display[self.cursor][1]
        return v, len(v)

    def nav_up(self, v, i):
        'Move cursor up one item, entering nav mode if needed.'
        if self.cursor == -1:
            self.enter_nav(v)
            if not self.display: return v, i
        else:
            self.cursor -= 1
            self.bounds()
        return self.sel()

    def nav_down(self, v, i):
        'Move cursor down one item, or restore original text at bottom.'
        if not self.display or self.cursor == -1: return v, i
        if self.cursor >= len(self.display) - 1:
            self.cursor = -1  # restore original typed text
            v = self.orig
            return v, len(v)
        self.cursor += 1
        self.bounds()
        return self.sel()

    def nav_pgup(self, v, i):
        'Scroll up one page.'
        if self.cursor == -1:
            self.enter_nav(v)
            if not self.display: return v, i
        nv = self.nvis()
        old_top = self.top
        self.cursor -= nv - 1
        self.top = old_top - nv + 1      # bottom = old top
        self.bounds()
        return self.sel()

    def nav_pgdn(self, v, i):
        'Scroll down one page.'
        if not self.display or self.cursor == -1: return v, i
        nv = self.nvis()
        old_bottom = self.top + nv - 1
        self.cursor += nv - 1
        self.top = old_bottom             # top = old bottom
        self.bounds()
        return self.sel()

    def nav_home(self, v, i):
        'Jump to first item, or move text cursor to start if not navigating.'
        if self.cursor == -1: return v, 0  # text cursor to start of line
        self.cursor = 0
        self.top = 0
        return self.sel()

    def nav_end(self, v, i):
        'Jump to last item, or move text cursor to end if not navigating.'
        if self.cursor == -1: return v, len(v)  # text cursor to end of line
        self.cursor = len(self.display) - 1
        self.bounds()
        return self.sel()

    def draw(self, value):
        'Update display list from fuzzy match and draw visible palette rows.'
        self.updater(value)
        scr = self.sheet._scr
        if not scr: return
        nv = self.nvis()
        w = min(100, self.sheet.windowWidth)

        if self.cursor >= 0 and value != self.display[self.cursor][1]:
            self.cursor = -1  # user typed something

        if self.cursor == -1:
            return  # don't draw until user navigates; display recomputed in enter_nav

        if not self.display: return
        ndisplay = min(len(self.display), nv)
        vis_start = max(0, min(self.top, len(self.display) - ndisplay))
        self.top = vis_start
        highlight = self.cursor - vis_start

        visible = self.display[vis_start:vis_start + ndisplay]
        if not visible: return
        box_y = self.sheet.windowHeight - ndisplay - self.input_height - 1
        if box_y < 0: return
        pal_cattr = colors.get_color('color_cmdpalette')
        vd.drawBox(scr, 0, box_y, w, ndisplay + 1, pal_cattr, bottom=False)
        for idx, (text, raw) in enumerate(visible):
            if idx == highlight:
                clipdraw(scr, box_y + 1 + idx, 1, f'> {text}', colors.color_menu_spec, w=w - 2)
            else:
                clipdraw(scr, box_y + 1 + idx, 1, f'  {text}', colors.color_cmdpalette, w=w - 2)


@VisiData.around
def input(vd, orig_input, *args, **kwargs):
    'Wrap vd.input() to show history palette when input has previous history.'
    _input_rows = kwargs.pop('_input_rows', 0) or 0
    _history_palette = kwargs.pop('_history_palette', True)
    type = kwargs.get('type')
    history = kwargs.get('history', [])
    updater = kwargs.get('updater', lambda v: None)
    bindings = kwargs.get('bindings', {})

    if not history and type:
        history = list(vd.inputHistory.setdefault(type, {}).keys())
        kwargs['history'] = history

    if not _history_palette \
            or not type \
            or not history \
            or not vd.cursesEnabled \
            or not vd.wantsHelp('cmdpalette'):
        return orig_input(vd, *args, **kwargs)

    sheet = vd.activeSheet
    items = list(history)   # 0=oldest .. N-1=newest
    pal = HistoryPalette(sheet, items, max(_input_rows, 1), updater)

    pal_bindings = dict(bindings)
    pal_bindings['Up'] = pal.nav_up
    pal_bindings['Down'] = pal.nav_down
    pal_bindings['PgUp'] = pal.nav_pgup
    pal_bindings['PgDn'] = pal.nav_pgdn
    pal_bindings['Home'] = pal.nav_home
    pal_bindings['End'] = pal.nav_end

    kwargs['updater'] = pal.draw
    kwargs['bindings'] = pal_bindings
    return orig_input(vd, *args, **kwargs)



"""Additional vim-like bindings for Visidata

Enable in .visidatarc with `import visidata.experimental.vimcompat`.
"""

from visidata import BaseSheet

# Jump to first row and column with 'go'
BaseSheet.bindkey('go', 'go-home')

# standard half-page movement
BaseSheet.bindkey('Ctrl+D', 'go-pagedown-half')
BaseSheet.bindkey('Ctrl+U', 'go-pageup-half')



import pytest
import warnings
from unittest.mock import Mock, patch

import visidata


class TestEditText:
    @pytest.fixture(autouse=True, scope='function')
    def setUp(self):
        self.chars = []
        visidata.vd.getkeystroke = Mock(side_effect=self.chars)
        visidata.vd.warning = warnings.warn

    @pytest.mark.parametrize('keys, result, kwargs', [
        ('Enter', '', {}),
        ('a b Home c d Ctrl+A e f Enter', 'efcdab', {}),
        ('a b Left 1 Left Left Left 2 Enter', '2a1b', {}), # Left, past home
        ('a b Ctrl+C', None, dict(exception=visidata.EscapeException)),
        ('a b Esc', None, dict(exception=visidata.EscapeException)),
        ('a Del Enter', 'a', {}),
        ('a b Left Del Enter', 'a', {}),
        ('a b Left c End d Enter', 'acbd', {}),
        ('a b Home Right c Enter', 'acb', {}),
        ('a b Bksp c Enter', 'ac', {}),

        # Backspace no longer deletes the first character at the start
        ('a b Home Bksp c Enter', 'cab', {}),

        # Backspace works in different combos, including on the mac.
        ('a b c Bksp Ctrl+H Left Del Enter', '', {}),

        ('a b c Ctrl+B Ctrl+B Ctrl+K Enter', 'a', {}),

        ('a Ctrl+R Enter', '', {}),
        ('a Ctrl+R Enter', 'foo', dict(value='foo')),

        # Two characters swaps characters
        ('a b Ctrl+T Enter', 'ba', {}),

        # Home with multiple characters acts like delete
        ('a b Home Ctrl+T Enter', 'b', {}),

        ('a b Left Ctrl+U Enter', 'b', {}),
        ('a b Ctrl+U c Enter', 'c', {}),

        ('w e Space a r e Space t h e Space w o r l d Ctrl+Left Ctrl+Left Ctrl+W Enter', 'we the world', {}),
        ('w e Space a r e Space t h e Space w o r l d Ctrl+Left Ctrl+Left Ctrl+Del Enter', 'we are world', {}),
    ])
    def test_keys(self, mock_screen, keys, result, kwargs):
        self.chars.extend([(' ' if k == 'Space' else k) for k in keys.split()])

        exception = kwargs.pop('exception', None)
        widget = visidata.InputWidget(**kwargs)
        if exception:
            with pytest.raises(exception):
                widget.editline(mock_screen, 0, 0, 0, attr=visidata.ColorAttr())
        else:
            try:
                r = widget.editline(mock_screen, 0, 0, 0, attr=visidata.ColorAttr())
            except StopIteration:
                assert False, "need Enter (or cancel) at end of keypresses"
            assert r == result



from visidata import VisiData, BaseSheet, vd

vd.theme_option('disp_splitwin_pct', 0, 'height of second sheet on screen')

vd.activePane = 1   # pane numbering starts at 1; pane 0 means active pane
BaseSheet.init('pane', lambda: 1)


@BaseSheet.api
def splitPane(sheet, pct=None):
    if vd.activeStack[1:]:
        undersheet = vd.activeStack[1]
        pane = 1 if undersheet.pane == 2 else 2
        vd.push(undersheet, pane=pane)
        vd.activePane = pane

    vd.options.disp_splitwin_pct = pct


@VisiData.api
def splitwin_close(vd):
    vd.options.disp_splitwin_pct = 0
    for vs in vd.stackedSheets:
        vs.pane = 1
        vd.activePane = 1


BaseSheet.addCommand('Z', 'splitwin-half', 'splitPane(vd.options.disp_splitwin_pct or 50)', 'ensure split pane is set and push under sheet onto other pane')
BaseSheet.addCommand('gZ', 'splitwin-close', 'vd.splitwin_close()', 'close split screen')
BaseSheet.addCommand('Tab', 'splitwin-swap', 'vd.activePane = 1 if sheet.pane == 2 else 2', 'jump to inactive pane')
BaseSheet.addCommand('gTab', 'splitwin-swap-pane', 'vd.options.disp_splitwin_pct=-vd.options.disp_splitwin_pct', 'swap panes onscreen')
BaseSheet.addCommand('zZ', 'splitwin-input', 'vd.options.disp_splitwin_pct = input("% height for split window: ", value=vd.options.disp_splitwin_pct)', 'set split pane to specific size')

vd.addMenuItems('''
    View > Split pane > in half > splitwin-half
    View > Split pane > in percent > splitwin-input
    View > Split pane > unsplit > splitwin-close
    View > Split pane > swap panes > splitwin-swap-pane
    View > Split pane > goto other pane > splitwin-swap
''')



import pytest
from unittest.mock import Mock, call

import visidata
from visidata import iterchars


class TestClipText:
    @pytest.mark.parametrize('s, dispw', [
        ('abcdef', 6),
        ('桜 高橋', 7),
        ('[:onclick sidebar-toggle][:reverse] b to toggle sidebar [:]', 21),
    ])
    def test_dispwidth(self, s, dispw):
        assert visidata.dispwidth(s) == dispw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' jso…', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcd…', 5),
        ('a', 1, 'a', 1),
        ('ab', 1, '…', 1),
        ('abc', 2, 'a…', 2),
        ('で', 1, '…', 1),
        ('でで', 1, '…', 1),
        ('でで', 2, '…', 1),
        ('でで', 3, 'で…', 3),
        ('ででで', 4, 'で…', 3),
        ('ででで', 5, 'でで…', 5),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
        (iterchars([1,2,3]), 4, '[3]…', 4),
        (iterchars([1,2,3]), 7, '[3] 1;…', 7),
        (iterchars([1,2,3]), 9, '[3] 1; 2…', 9),
        (iterchars([1,2,3]), 11, '[3] 1; 2; 3', 11),
        (iterchars([1,2,3]), 12, '[3] 1; 2; 3', 11),
        (iterchars({'a':1, 'b':2, 'c':3}), 7, '{3} a=…', 7),
        (iterchars({'a':1, 'b':2, 'c':3}), 15, '{3} a=1 b=2 c=3', 15),
        (iterchars({'a':1, 'b':2, 'で':3}), 13, '{3} a=1 b=2 …', 13),
        (iterchars({'a':1, 'b':2, 'で':3}), 16, '{3} a=1 b=2 で=3', 16),
    ])
    def test_clipstr(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' jsあ', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcあ', 5),
        ('a', 1, 'a', 1),
        ('ab', 1, 'a', 1),
        ('abc', 2, 'あ', 2),
        ('で', 1, '', 0),
        ('でで', 1, '', 0),
        ('でで', 2, 'あ', 2),
        ('でで', 3, 'あ', 2),
        ('ででで', 4, 'であ', 4),
        ('ででで', 5, 'であ', 4),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
        (iterchars([1,2,3]), 1, '[', 1),
        (iterchars({'a':1, 'b':2, 'c':3}), 1, '{', 1),
    ])
    def test_clipstr_wide_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator='あ')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' json', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcd', 4),
        ('a', 1, 'a', 1),
        ('ab', 1, 'a', 1),
        ('abc', 2, 'ab', 2),
        ('で', 1, '', 0),
        ('でで', 1, '', 0),
        ('でで', 2, 'で', 2),
        ('でで', 3, 'で', 2),
        ('ででで', 4, 'でで', 4),
        ('ででで', 5, 'でで', 4),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
    ])
    def test_clipstr_empty_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator='')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, truncator, clippeds, clippedw', [
        ('first\nsecond\n\nthird\n\n\n', 22, '',  'first·second··third···', 22),
        ('first\nsecond\n\nthird\n\n\n', 22, '…', 'first·second··third···', 22),
        ('first\nsecond\n\nthird\n\n\n', 21, '',  'first·second··third··', 21),
        ('first\nsecond\n\nthird\n\n\n', 21, '…', 'first·second··third·…', 21),
        (''.join([chr(i) for i in range(256)]), 256, '',
            '································ !"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~··································¡¢£¤¥¦§¨©ª«¬\xad®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ', 256),
    ])
    def test_clipstr_unprintable(self, s, w, truncator, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator=truncator, oddspace='·')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, 'jsonl', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'bcdで', 5),
        ('でbcdで', 6, 'bcdで', 5),
        ('でbcdefghiで', 10, 'bcdefghiで', 10),
        ('でbcdefghiで', 3, 'iで', 3),
        ('でbcdで', 2, 'で', 2),
        ('でbcdで', 1, '', 0),
        ('でbcdで', 0, '', 0),
        ('でbcdで', -1, '', 0),
    ])
    def test_clipstr_start(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_start(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('aAbcで', 6, 'aAbcで', 6),
        ('aAbcで', 7, 'aAbcで', 6),
        ('aAbcで', 1000, 'aAbcで', 6),
        ('aAbcで', 5, '…bcで', 5),
        ('でbcで', 5, '…bcで', 5),
        ('でででででbcで', 5, '…bcで', 5),
        ('でbcで', 3, '…で', 3),
        ('でbcで', 2, '…', 1),
        ('でbcで', 1, '…', 1),
        ('でbcで', 0, '', 0),
        ('でbcで', -1, '', 0),
    ])
    def test_clipstr_start_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_start(s, w, truncator='…')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('1234567890', 6, '12…890', 6),
        ('1234567890', 7, '123…890', 7),
        ('1234567890', 8, '123…7890', 8),
        ('1234567890', 9, '1234…7890', 9),
        ('1234567890', 10, '1234567890', 10),
        ('1234567890', 11, '1234567890', 10),
        ('1234567890', 99, '1234567890', 10),
        # all full-width characters
        ('ででででで', 0,  '', 0),
        ('ででででで', 1,  '…', 1),
        ('ででででで', 2,  '…', 1),
        ('ででででで', 3,  '…で', 3),
        ('ででででで', 4,  '…で', 3),
        ('ででででで', 5,  'で…で', 5),
        ('ででででで', 6,  'で…で', 5),
        ('ででででで', 7,  'で…でで', 7),
        ('ででででで', 8,  'で…でで', 7),
        ('ででででで', 9,  'でで…でで', 9),
        ('ででででで', 10, 'ででででで', 10),
        ('ででででで', 11, 'ででででで', 10),
        ('ででででで', 99, 'ででででで', 10),
        # odd string length, with mix of full-width characters
        ('ででaaでa', 0,  '', 0),
        ('ででaaでa', 1,  '…', 1),
        ('ででaaでa', 2,  '…a', 2),
        ('ででaaででa', 3,  '…a', 2),
        ('ででaaででa', 4,  '…でa', 4),
        ('ででaaででa', 5,  'で…a', 4),
        ('ででaaででa', 6,  'で…でa', 6),
    ])
    def test_clipstr_middle(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_middle(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, dispw, clipped', [
        #clip front half
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]…[:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),
        #clip front half, dropping escaped markup
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]\uFFF9[:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:]\uFFFB[:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]…[:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),

        #clip back half, when front half is an exact fit at 24 wide
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234…[:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),

        #clip front half, when front half reaches 25 wide
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]12345[:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]…[:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),

        #clip back half, when front half is 24 wide and back half is 27 wide
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]1234567[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234…[:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]'),

        #no clipping for a string exactly 50 wide, when front is 24 wide and back is 26
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        50,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]'),

        #trimming to very short widths
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        5,
        '…'),
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        1,
        '…'),
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        0,
        ''),
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:]1234[:menu-active]123456[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:]',
        -1,
        ''),
        ('[:onclick jump-sheet-1]a[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:onclick jump-sheet-4]b[:]',
        4,
        '[:onclick jump-sheet-1]a[:]…[:onclick jump-sheet-4]b[:]'),
        ('[:onclick jump-sheet-1]a[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:onclick jump-sheet-4]b[:]',
        2,
        '…[:onclick jump-sheet-4]b[:]'),

        #no clipping, contents have plenty of room
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        100,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),

        #no clipping, contents fit exactly
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        70,
        '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),

        #contents are too wide by 1
        ('[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:][:menu-active]ten_chars4[:][:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]',
        69,
         '[:onclick jump-sheet-1]ten_chars0[:][:onclick jump-sheet-2]ten_chars1[:][:menu-active]ten_chars3[:]…[:menu-active]ten_chars5[:][:menu-active]ten_chars6[:][:menu-active]ten_chars7[:]'),
    ])
    def test_truncate_markup_middle(self, s, dispw, clipped):
        output = visidata.clip_markup_middle(s, dispw)
        assert output == clipped

    @pytest.mark.parametrize('text, width, expected', [
        # color tag spanning two lines should carry over
        ('[:error]line one\nline two[/]', 80,
         [('[:error]line one[/]', 'line one'),
          ('[:error]line two[/]', 'line two')]),
        # nested tags spanning lines
        ('[:bold]a\n[:error]b[/]\nc[/]', 80,
         [('[:bold]a[/]', 'a'),
          ('[:bold][:error]b[/][/]', 'b'),
          ('[:bold]c[/]', 'c')]),
        # no markup, multiline (should work as before)
        ('hello\nworld', 80,
         [('hello', 'hello'),
          ('world', 'world')]),
        # single line with markup (should work as before)
        ('[:error]oops[/]', 80,
         [('[:error]oops[/]', 'oops')]),
        # nested tags spanning lines, closed with [:]
        ('[:bold]a\n[:error]b\nc[:]\nd [:bold]e[/]', 80,
         [('[:bold]a[/]', 'a'),
          ('[:bold][:error]b[/][/]', 'b'),
          ('[:bold][:error]c[:]', 'c'),
          ('d [:bold]e[/]', 'd e'),
         ]),
    ])
    def test_wraptext_color_spans_lines(self, text, width, expected):
        result = list(visidata.wraptext(text, width=width))
        assert result == expected

    @pytest.mark.parametrize('text, width, expected', [
        # escaped plain text lacking markup, to demonstrate
        ('\uFFF9.:bold.a.:.123\uFFFb', 10,
         [('\uFFF9.:bold.a.:.123\uFFFb', '.:bold.a.:'),
          ('', '.123')]),
        # escaped markup wraps just like the example above
        ('\uFFF9[:bold]a[:]123\uFFFb', 10,
         [('\uFFF9[:bold]a[:]123\uFFFb', '[:bold]a[:'),
          ('', ']123')]),
    ])
    def test_wraptext_escaped_literals(self, text, width, expected):
        result = list(visidata.wraptext(text, width=width))
        assert result == expected



'''
# LLM Analysis Feature

Analyze column data using Anthropic Claude LLM with natural language queries.

## Usage

1. Position cursor on a column containing text data
2. Run `llm-analyze-column` command (or `llm-analyze-selected` for selected rows only)
3. Enter your analysis query (e.g. "Did the model exhibit reasoning?")
4. Four new columns will be added with results:
   - `{colname}_llm_score` (0-1 float score)
   - `{colname}_llm_analysis` (text explanation)
   - `{colname}_llm_raw` (hidden raw JSON response)
   - `{colname}_llm_tokens` (token count for cost tracking)

## Configuration

Set in ~/.visidatarc or via environment variables:

    options.llm_model = "claude-haiku-4-5"
    options.llm_max_tokens = 1024
    options.llm_temperature = 1.0
    options.llm_rate_delay = 0.5  # seconds between requests
    options.llm_template_path = ""  # path to custom template file

    export ANTHROPIC_API_KEY="..."
## Custom Templates

Create a text file with {query} and {text} placeholders:

    Analyze this text: {text}
    Question: {query}
    Return JSON: {{"score": 0-1, "analysis": "explanation"}}

Then set: `options.llm_template_path = "/path/to/template.txt"`
'''

import os
import time
import json
import re

from visidata import vd, VisiData, Sheet, SettableColumn, asyncthread, Progress

# Configuration options
vd.option('llm_model', 'claude-haiku-4-5', 'Anthropic model name')
vd.option('llm_max_tokens', 1024, 'max tokens for LLM response')
vd.option('llm_temperature', 1.0, 'temperature for LLM sampling (0-1)')
vd.option('llm_rate_delay', 0.1, 'delay in seconds between API requests')
vd.option('llm_template_path', '', 'path to custom template file (empty = use default)')

# Default template with structured JSON output
DEFAULT_TEMPLATE = '''Analyze the following text and respond in JSON format.

Query: {query}
Text: {text}

Provide:
1. A score from 0 to 1 indicating how well the text matches or satisfies the query
2. A brief analysis explaining your reasoning

Response format: {{"score": <float 0-1>, "analysis": "<text>"}}'''


@VisiData.lazy_property
def anthropic_client(vd):
    'Initialize and return Anthropic API client.'
    anthropic = vd.importExternal('anthropic')
    api_key = os.environ.get('ANTHROPIC_API_KEY') or vd.fail('set $ANTHROPIC_API_KEY (Get API key from https://console.anthropic.com/')
    return anthropic.Anthropic(api_key=api_key)


@VisiData.api
def loadLLMTemplate(vd):
    'Load LLM template from file or return default template.'
    template_path = vd.options.llm_template_path
    if template_path:
        try:
            with open(template_path) as f:
                return f.read()
        except Exception as e:
            vd.exceptionCaught(e)
            vd.fail(f'Could not load template from {template_path}')

    return DEFAULT_TEMPLATE


@VisiData.api
def queryLLM(vd, prompt):
    'Query Anthropic LLM with prompt and return parsed response.'
    response = vd.anthropic_client.messages.create(
        model=vd.options.llm_model,
        max_tokens=vd.options.llm_max_tokens,
        temperature=vd.options.llm_temperature,
        messages=[{'role': 'user', 'content': prompt}]
    )

    raw_text = response.content[0].text
    tokens = response.usage.input_tokens + response.usage.output_tokens

    # Try to parse JSON response
    try:
        data = json.loads(raw_text)
        score = float(data.get('score', 0))
        analysis = str(data.get('analysis', ''))
    except (json.JSONDecodeError, ValueError, KeyError):
        # Fallback: try to extract score with regex
        score_match = re.search(r'"?score"?\s*:\s*(\d+\.?\d*)', raw_text)
        score = float(score_match.group(1)) if score_match else None
        analysis = raw_text  # use raw response as analysis

    return {
        'score': score,
        'analysis': analysis,
        'raw': raw_text,
        'tokens': tokens
    }


@Sheet.api
@asyncthread
def llmAnalyzeColumn(sheet, col, rows, query):
    'Analyze column data with LLM using given query.'
    # Load template
    template = vd.loadLLMTemplate()

    # Warn about large datasets
    n_rows = len(rows)
    if n_rows > 100:
        if not vd.confirm(f'Analyze {n_rows} rows? This may take some time and money.'):
            return

    # Create result columns
    score_col = SettableColumn(f'{col.name}_llm_score', type=float)
    analysis_col = SettableColumn(f'{col.name}_llm_analysis', type=str, width=40)
    raw_col = SettableColumn(f'{col.name}_llm_raw', type=str, width=0)  # hidden
    tokens_col = SettableColumn(f'{col.name}_llm_tokens', type=int, width=8)

    sheet.addColumnAtCursor(score_col, analysis_col, raw_col, tokens_col)

    # Track statistics
    processed = 0
    skipped = 0
    errors = 0
    total_tokens = 0

    # Process rows sequentially
    for row in Progress(rows, gerund='analyzing with LLM'):
        cell_text = col.format(col.getTypedValue(row))

        # Skip empty cells
        if not cell_text or str(cell_text).strip() == '':
            skipped += 1
            continue

        # Truncate if too long (rough estimate: 4 chars ~= 1 token)
        max_chars = vd.options.llm_max_tokens * 3
        if len(cell_text) > max_chars:
            cell_text = cell_text[:max_chars] + '...'
            vd.warning(f'Cell content truncated (exceeded {max_chars} chars)')

        # Build prompt
        prompt = template.format(query=query, text=cell_text)

        # Call API with error handling
        try:
            result = vd.queryLLM(prompt)

            score_col.setValue(row, result['score'])
            analysis_col.setValue(row, result['analysis'])
            raw_col.setValue(row, result['raw'])
            tokens_col.setValue(row, result['tokens'])

            total_tokens += result['tokens']
            processed += 1

        except Exception as e:
            vd.exceptionCaught(e)
            analysis_col.setValue(row, f'Error: {e}')
            errors += 1

        # Rate limiting
        time.sleep(vd.options.llm_rate_delay)

    # Report results
    msg = f'LLM analysis complete: {processed} rows processed'
    if skipped > 0:
        msg += f', {skipped} skipped (empty)'
    if errors > 0:
        msg += f', {errors} errors'
    if total_tokens > 0:
        msg += f' ({total_tokens} tokens)'

    vd.status(msg)


# Command bindings - no default keybindings, users can add their own
Sheet.addCommand('', 'llm-analyze-column',
    'llmAnalyzeColumn(cursorCol, rows, vd.input("LLM query: ", type="llm-query"))',
    'analyze all rows in current column with LLM')

Sheet.addCommand('', 'llm-analyze-selected',
    'llmAnalyzeColumn(cursorCol, someSelectedRows, vd.input("LLM query: ", type="llm-query"))',
    'analyze selected rows only with LLM')

# Menu integration
vd.addMenuItems('''
    Column > Analyze with LLM > all rows > llm-analyze-column
    Column > Analyze with LLM > selected rows > llm-analyze-selected
''')



import re
import time

from visidata import vd, BaseSheet, Path, VisiData

vd.replay_output_path = ''  # output path set by replay-reset, used by replay-output
vd.replay_line = 0  # current line number within a test
vd.replay_start_time = 0  # timestamp when current test started
vd.replay_allowed_errors = []  # set by allow-error, checked by status hook


@VisiData.before
def replayOne(vd, r):
    vd.replay_line += 1


@BaseSheet.api
def replay_reset(vs):  # noqa: ARG001
    'Initialize state for a new test.'
    p = vd.input("output path: ")
    vd.resetVisiData()
    vd.options.batch = True
    vd.replay_output_path = p
    vd.replay_line = 0
    vd.replay_start_time = time.time()
    vd.replay_allowed_errors = []
    vd.statusHistory.clear()  # clear for cross-test isolation


@BaseSheet.api
def replay_end(vs):  # noqa: ARG001
    'Reset state for next test (no output).'
    if vd.options.debug:
        print(f'{time.time() - vd.replay_start_time:.1f}s  {vd.replay_output_path}')

_default_printStatus = VisiData.printStatus

@VisiData.api
def printStatus(vd, *args, priority=0, source=None):
    'Print labeled status during bulk replay; default stderr otherwise.'
    if priority > 0 and vd.replay_output_path:
        msg = str(args[0])
        allowed = any(re.search(p, msg) for p in vd.replay_allowed_errors)
        if not allowed:
            print(f'{vd.replay_output_path}:{vd.replay_line}: {msg}')
        elif vd.options.debug:
            print(f'{vd.replay_output_path}:{vd.replay_line}: {msg} (expected)')
    else:
        _default_printStatus(vd, *args, priority=priority, source=source)

@BaseSheet.api
def replay_output(vs):
    'Save current sheet to path from replay-reset and reset state for next test.'
    outpath = Path(vd.replay_output_path)
    vd.saveSheets(outpath, vs, confirm_overwrite=False)
    vd.sync()
    if vd.options.debug:
        print(f'{time.time() - vd.replay_start_time:.1f}s  {vd.replay_output_path}')


@BaseSheet.api
def allow_error(vs, pattern:str):  # noqa: ARG001
    'suppress expected error messages matching regex pattern during batch replay'
    vd.replay_allowed_errors.append(pattern)


@BaseSheet.api
def replay_exit(vs):  # noqa: ARG001
    'Exit cleanly at end of batch replay.'
    vd.sheets.clear()


BaseSheet.addCommand('', 'replay-reset', 'replay_reset()', 'initialize test state for a test', testable=False)
BaseSheet.addCommand('', 'replay-end', 'replay_end()', 'reset state for test runner (no output)', testable=False)
BaseSheet.addCommand('', 'replay-output', 'replay_output()', 'save output and reset state for test runner', testable=False)
BaseSheet.addCommand('', 'replay-exit', 'replay_exit()', 'exit cleanly at end of batch replay', testable=False)
BaseSheet.addCommand('', 'allow-error', 'allow_error(vd.input("allow error: "))', 'suppress expected error messages matching regex pattern', testable=False)



import json

from visidata import Sheet, VisiData, TypedWrapper, anytype, date, vlen, Column, vd, asyncthread, Progress, InvertedCanvas
from collections import defaultdict


@VisiData.api
def open_parquet(vd, p):
    return ParquetSheet(p.base_stem, source=p)


class ParquetColumn(Column):
    @property
    def readonly(self) -> bool:
        return False

    def calcValue(self, row):
        if self.name in row:  #2890
            return row[self.name]
        rownum = row.get("__rownum__")
        if rownum is None:
            return None
        val = self.source[rownum]
        if val.type == 'large_string':
            buf = val.as_buffer()
            if buf is None:
                return None
            return memoryview(buf)[:2**20].tobytes().decode('utf-8')
        else:
            return val.as_py()

    def putValue(self, row, val):
        row[self.name] = val


class GeometryColumn(ParquetColumn):
    'Parquet column containing WKB-encoded geometries (GeoParquet).'
    @property
    def readonly(self) -> bool:
        return True

    def calcValue(self, row):
        val = super().calcValue(row)
        if val is None:
            return None
        return vd.importExternal('shapely').from_wkb(val)

    def formatValue(self, typedval, width=None):
        if typedval is None:
            return None
        shapely = vd.importExternal('shapely')
        n = shapely.get_num_coordinates(typedval)
        return f'{typedval.geom_type}[{n}]'


def _geoparquet_columns(schema):
    'Return set of column names that are WKB geometries per GeoParquet metadata.'
    names = set()
    meta = schema.metadata or {}
    geo = meta.get(b'geo')
    if geo:
        try:
            names.update(json.loads(geo).get('columns', {}).keys())
        except Exception as e:
            vd.exceptionCaught(e)
    # fallback: per-field ARROW extension marker
    for i in range(len(schema)):
        field = schema.field(i)
        fmd = field.metadata or {}
        if fmd.get(b'ARROW:extension:name') == b'geoarrow.wkb':
            names.add(field.name)
    return names


class ParquetSheet(Sheet):
    # rowdef: {'__rownum__':int, parquet_col:overridden_value, ...}
    def iterload(self):
        pa = vd.importExternal("pyarrow", "pyarrow")
        pq = vd.importExternal("pyarrow.parquet", "pyarrow")
        from visidata.loaders.arrow import arrow_to_vdtype

        if self.source.is_dir():
            self.tbl = pq.read_table(str(self.source))
        else:
            with self.source.open('rb') as f:
                self.tbl = pq.read_table(f)

        geocols = _geoparquet_columns(self.tbl.schema)

        self.columns = []
        for colname, col in zip(self.tbl.column_names, self.tbl.columns):
            if colname in geocols:
                c = GeometryColumn(colname, type=anytype, source=col, cache=True)
            else:
                c = ParquetColumn(colname,
                                  type=arrow_to_vdtype(col.type),
                                  source=col,
                                  cache=(col.type.id == pa.lib.Type_LARGE_STRING))
            self.addColumn(c)

        for i in range(self.tbl.num_rows):
            yield dict(__rownum__=i)

    def geometryColumn(self):
        'Return first GeometryColumn on this sheet, or None.'
        for c in self.columns:
            if isinstance(c, GeometryColumn):
                return c
        return None


class ParquetGeoCanvas(InvertedCanvas):
    aspectRatio = 1.0

    @asyncthread
    def reload(self):
        self.reset()
        geocol = self.source.geometryColumn()
        if geocol is None:
            vd.warning('no geometry column')
            return
        for row in Progress(self.sourceRows):
            g = geocol.getTypedValue(row)
            if g is None:
                continue
            self._plot_geom(g, self.plotColor(self.source.rowkey(row)), row)
        self.refresh()

    def _plot_geom(self, g, attr, row):
        t = g.geom_type
        if t == 'Point':
            self.point(g.x, g.y, attr, row)
            disptext = self.textCol.getDisplayValue(row)
            if disptext:
                self.label(g.x, g.y, disptext, attr, row)
        elif t in ('LineString', 'LinearRing'):
            self.polyline(list(g.coords), attr, row)
        elif t == 'Polygon':
            self.polyline(list(g.exterior.coords), attr, row)
            for ring in g.interiors:
                self.polyline(list(ring.coords), attr, row)
        elif t in ('MultiPoint', 'MultiLineString', 'MultiPolygon', 'GeometryCollection'):
            for sub in g.geoms:
                self._plot_geom(sub, attr, row)
        else:
            vd.warning(f'unsupported geometry type `{t}`')


@VisiData.api
def save_parquet(vd, p, sheet):
    pa = vd.importExternal("pyarrow")
    pq = vd.importExternal("pyarrow.parquet", "pyarrow")

    typemap = {
        anytype: pa.string(),
        int: pa.int64(),
        vlen: pa.int64(),
        float: pa.float64(),
        str: pa.string(),
        date: pa.date64(),
        # list: pa.array(),
    }

    for t in vd.numericTypes:
        if t not in typemap:
            typemap[t] = pa.float64()

    databycol = defaultdict(list)  # col -> [values]

    for typedvals in sheet.iterdispvals(format=False):
        for col, val in typedvals.items():
            if isinstance(val, TypedWrapper):
                val = None

            databycol[col].append(val)

    data = [
        pa.array(vals, type=typemap.get(col.type, pa.string()))
        for col, vals in databycol.items()
    ]

    schema = pa.schema(
        [(c.name, typemap.get(c.type, pa.string())) for c in sheet.visibleCols]
    )
    with p.open_bytes(mode="w") as outf:
        with pq.ParquetWriter(outf, schema) as writer:
            writer.write_batch(
                pa.record_batch(data, names=[c.name for c in sheet.visibleCols])
            )


ParquetSheet.addCommand('.', 'plot-row', 'vd.push(ParquetGeoCanvas(name+"_map", source=sheet, sourceRows=[cursorRow], textCol=cursorCol))', 'plot geometry in current row')
ParquetSheet.addCommand('g.', 'plot-rows', 'vd.push(ParquetGeoCanvas(name+"_map", source=sheet, sourceRows=rows, textCol=cursorCol))', 'plot geometries in all rows')

vd.addGlobals(ParquetGeoCanvas=ParquetGeoCanvas)



import curses

from visidata import vd, VisiData, BaseSheet, Sheet, AttrDict, dispwidth


# registry of mouse events.  cleared before every draw cycle.
vd.mousereg = []  # list of AttrDict(y=, x=, h=, w=, buttonfuncs=dict)

# sheet mouse position for current mouse event
BaseSheet.init('mouseX', int)
BaseSheet.init('mouseY', int)
# sheet info for a click that may become a drag
BaseSheet.init('drag1')


@VisiData.after
def initCurses(vd):
    if not getattr(curses, 'mousemask', None):
      return
    curses.MOUSE_ALL = 0xffffffff
    curses.mouseEvents = {}

    if not vd.enableMouse(bool(vd.options.mouse_interval)):
        return

    curses.def_prog_mode()
    curses.mouseinterval(vd.options.mouse_interval)

    for k in dir(curses):
        if k.startswith('BUTTON') or k in ('REPORT_MOUSE_POSITION', '2097152'):
            curses.mouseEvents[getattr(curses, k)] = k


@VisiData.api
def enableMouse(vd, b:bool) -> bool:  #2913 #2851
    'Call curses.mousemask(all if b else 0) only if available.  Return True if mouse enabled.'
    if not hasattr(curses, 'mousemask'):
        return False
    mm, _ = curses.mousemask(getattr(curses, 'MOUSE_ALL', 0xffffffff) if b else 0)
    return bool(mm)


@VisiData.after
def clearCaches(vd):
    vd.mousereg = []


@VisiData.api
def onMouse(vd, scr, x, y, w, h, **kwargs):
    px, py = vd.getrootxy(scr)
    e = AttrDict(x=x+px, y=y+py, w=w, h=h, buttonfuncs=kwargs)
    vd.mousereg.append(e)


@VisiData.api
def getMouse(vd, _x, _y, button):
    for reg in vd.mousereg[::-1]:
        if reg.x <= _x < reg.x+reg.w and reg.y <= _y < reg.y+reg.h and button in reg.buttonfuncs:
            return reg.buttonfuncs[button]


@VisiData.api
def parseMouse(vd, **kwargs):
    'Return list of mouse interactions (clicktype, y, x, name, scr) for curses screens given in kwargs as name:scr.'

    devid, x, y, z, bstate = curses.getmouse()

    clicktype = ''
    if bstate & curses.BUTTON_CTRL:
        clicktype += "Ctrl+"
        bstate &= ~curses.BUTTON_CTRL
    if bstate & curses.BUTTON_ALT:
        clicktype += "Alt+"
        bstate &= ~curses.BUTTON_ALT
    if bstate & curses.BUTTON_SHIFT:
        clicktype += "Shift+"
        bstate &= ~curses.BUTTON_SHIFT

    keystroke = clicktype + curses.mouseEvents.get(bstate, str(bstate))
    ret = AttrDict(keystroke=keystroke, y=y, x=x, found=[])
    for winname, winscr in kwargs.items():
        if not winscr:
            continue
        px, py = vd.getrootxy(winscr)
        mh, mw = winscr.getmaxyx()
        if py <= y < py+mh and px <= x < px+mw:
            ret.found.append(winname)
#            vd.debug(f'{keystroke} at ({x-px}, {y-py}) in window {winname} {winscr}')

    return ret


@VisiData.api
def handleMouse(vd, sheet):
    r = None
    try:
        vd.keystrokes = ''
        pct = vd.windowConfig['pct']
        topPaneActive = ((vd.activePane == 2 and pct < 0)  or (vd.activePane == 1 and pct > 0))
        bottomPaneActive = ((vd.activePane == 1 and pct < 0)  or (vd.activePane == 2 and pct > 0))
        r = vd.parseMouse(top=vd.winTop, bot=vd.winBottom, menu=vd.scrMenu)
        if (bottomPaneActive and 'top' in r.found) or (topPaneActive and 'bot' in r.found):
            vd.activePane = 1 if vd.activePane == 2 else 2
            sheet = vd.activeSheet

        f = vd.getMouse(r.x, r.y, r.keystroke)
        winx, winy = vd.getrootxy(sheet._scr)
        sheet.mouseX, sheet.mouseY = r.x-winx, r.y-winy
        if f:
            if isinstance(f, str):
                if f.startswith('onclick'):
                    if '://' in f:
                        vd.launchBrowser(f[8:])
                    else:
                        sheet.execCommand(f[8:])
                else:
                    for cmd in f.split():
                        sheet.execCommand(cmd)
            else:
                f(r.y, r.x, r.keystroke)

            vd.keystrokes = vd.prettykeys(r.keystroke)
            return ''  #  handled
    except curses.error:
        pass

    return r.keystroke if r else ''


@Sheet.api
def visibleColInfoAtX(sheet, x):
    '''return (vcolidx, is_separator) for the x-coordinate'''
    for vcolidx, (colx, w) in sheet._visibleColLayout.items():
        if vcolidx == sheet.nVisibleCols-1:
            sep = vd.options.disp_rowend_sep
        elif (sheet.keyCols and sheet.availCols[vcolidx] is sheet.keyCols[-1]):
            sep = vd.options.disp_keycol_sep
        else:
            sep = vd.options.disp_column_sep
        sepw = dispwidth(sep, literal=True)
        if colx <= x <= colx+w+sepw-1:
            if colx <= x <= colx+w-1:  #in the cell
                return (vcolidx, False)
            else:   #in the end-of-column separator
                return (vcolidx, True)
    return (None, False)


@Sheet.api
def visibleRowAtY(sheet, y):
    for rowidx, (rowy, h) in sheet._rowLayout.items():
        if rowy <= y <= rowy+h-1:
            return rowidx


Sheet.addCommand('BUTTON1_PRESSED', 'go-mouse-1', 'go_mouse(drag_button=1)', 'set cursor to row and column where mouse was clicked, or start a drag event')
Sheet.bindkey('BUTTON1_CLICKED', 'go-mouse-1')
Sheet.addCommand('BUTTON1_RELEASED', 'check-drag', 'drag_button1()', 'resize column if col separator has been dragged')
Sheet.addCommand('BUTTON3_PRESSED', 'go-mouse', 'go_mouse()', 'set cursor to row and column where mouse was clicked')
@Sheet.api
def go_mouse(sheet, drag_button=None):
    if sheet.mouseY == sheet.windowHeight-1:
        return
    ridx = sheet.visibleRowAtY(sheet.mouseY)
    if ridx is not None:
        sheet.cursorRowIndex = ridx
    cidx, is_column = sheet.visibleColInfoAtX(sheet.mouseX)
    if cidx is not None:
        sheet.cursorVisibleColIndex = cidx
        if drag_button == 1 and is_column:
            # save info for a possible drag event
            sheet.drag1 = (sheet.availCols[cidx], sheet.mouseX)
        else:
            sheet.drag1 = None

@Sheet.api
def drag_button1(sheet):
    if not sheet.drag1:
        return
    delta_w = sheet.mouseX - sheet.drag1[1]
    col = sheet.drag1[0]
    sheet.drag1 = None
    if col.width and col.width > 0:
        new_w = col.width + delta_w
        if new_w > 0:
            col.setWidth(new_w)

Sheet.addCommand(None, 'scroll-mouse', 'sheet.topRowIndex=cursorRowIndex-mouseY+1', 'scroll to mouse cursor location')

Sheet.addCommand('ScrollUp', 'scroll-up', 'cursorDown(options.scroll_incr); sheet.topRowIndex += options.scroll_incr', 'scroll up by one row increment')
Sheet.addCommand('ScrollDown', 'scroll-down', 'cursorDown(-options.scroll_incr); sheet.topRowIndex -= options.scroll_incr', 'scroll down by one row increment')
